"""Tests for csg/_compose.py"""

import numpy as np
import pandas as pd
import pytest
import xarray as xr

import primap2.csg
import primap2.csg._compose
from primap2.tests.csg.utils import get_single_ts


def assert_copied_from_input_data(
    filtered_result: xr.DataArray,
    filtered_initial: xr.DataArray,
    common_filter: dict[str, str],
):
    """Assert that timeseries in the result were copied from the input data."""
    expected = (
        filtered_initial.pr.loc[common_filter]
        .drop_vars("scenario (FAOSTAT)")
        .drop_vars("source")
        .expand_dims(dim={"source": ["composed"]})
    )
    result = filtered_result.pr.loc[common_filter]
    xr.testing.assert_identical(result, expected)


def test_compose_simple(opulent_ds):
    """A test with relatively complex input data but simple rules.

    The input data has many dimensions and we use two as priority dimensions. There
    are some NaN data points in one source.

    The input rules use two sources for all entities, plus one additional for CH4. For
    every entity, we use the simple SubstitutionStrategy for all sources.
    """
    input_data = opulent_ds
    input_data = input_data.drop_vars(["population", "SF6 (SARGWP100)"]).pr.loc[
        {"category": ["0", "1"]}
    ]
    input_data["CO2"].loc[{"source": "RAND2020", "time": ["2000", "2001"]}] = (
        np.nan * primap2.ureg("Mt CO2 / year")
    )
    # we now have dimensions time, area (ISO3), category (IPCC 2006), animal (FAOSTAT)
    # product (FAOSTAT), scenario (FAOSTAT), provenance, model, source
    # We have variables (entities): CO2, SF6, CH4
    # We have sources: RAND2020, RAND2021
    # We have scenarios: highpop, lowpop
    # Idea: we use source, scenario as priority dimensions, everything else are fixed
    # coordinates.

    # generally, prefer source RAND2020, scenario lowpop,
    # then use source RAND2021, scenario highpop.
    # however, for Columbia CH4, use source RAND2020, scenario highpop as highest priority
    # (this combination is not used at all otherwise).
    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source", "scenario (FAOSTAT)"],
        priorities=[
            {
                "entity": "CH4",
                "area (ISO3)": "COL",
                "source": "RAND2020",
                "scenario (FAOSTAT)": "highpop",
            },
            {"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"},
            {"source": "RAND2021", "scenario (FAOSTAT)": "highpop"},
        ],
    )
    # we use straight substitution always, but specify it somewhat fancy
    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            (
                {"entity": ent},
                primap2.csg.SubstitutionStrategy(),
            )
            for ent in input_data.keys()
        ]
    )

    result = primap2.csg.compose(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=strategy_definition,
    )
    # The caller of `compose` is responsible for re-adding priority dimensions
    # if necessary
    result = result.expand_dims(dim={"source": ["composed"]})
    result.pr.ensure_valid()
    assert "CO2" in result.keys()
    assert "Processing of CO2" in result.keys()
    assert_copied_from_input_data(
        result["CH4"],
        input_data["CH4"].loc[{"source": "RAND2020", "scenario (FAOSTAT)": "highpop"}],
        {"area": "COL"},
    )
    result_col_proc = (
        result["Processing of CH4"].loc[{"area (ISO3)": "COL"}].values.flat[0]
    )
    assert len(result_col_proc.steps) == 1
    assert result_col_proc.steps[0].time == "all"
    assert result_col_proc.steps[0].function == "substitution"
    assert "'source': 'RAND2020'" in result_col_proc.steps[0].description
    assert "'scenario (FAOSTAT)': 'highpop'" in result_col_proc.steps[0].description

    assert_copied_from_input_data(
        result["CH4"],
        input_data["CH4"].loc[{"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"}],
        {"area": "ARG"},
    )
    result_arg_proc = (
        result["Processing of CH4"].loc[{"area (ISO3)": "ARG"}].values.flat[0]
    )
    assert len(result_arg_proc.steps) == 1
    assert result_arg_proc.steps[0].time == "all"
    assert result_arg_proc.steps[0].function == "substitution"
    assert (
        result_arg_proc.steps[0].source
        == "{'source': 'RAND2020', 'scenario (FAOSTAT)': 'lowpop'}"
    )

    assert_copied_from_input_data(
        result["CO2"],
        input_data["CO2"].loc[{"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"}],
        {"area": "COL", "time": slice("2002", None)},
    )
    result_col_co2_proc = (
        result["Processing of CO2"].loc[{"area (ISO3)": "COL"}].values.flat[0]
    )
    assert len(result_col_co2_proc.steps) == 2
    assert result_col_co2_proc.steps[0].function == "substitution"
    np.testing.assert_array_equal(
        result_col_co2_proc.steps[1].time,
        np.array(["2000", "2001"], dtype=np.datetime64),
    )
    assert result_col_co2_proc.steps[1].function == "substitution"
    assert "'source': 'RAND2020'" in result_col_co2_proc.steps[0].description
    assert "'scenario (FAOSTAT)': 'lowpop'" in result_col_co2_proc.steps[0].description


def test_compose_exclude_result(opulent_ds):
    """In this test, we exclude parts of the result from processing."""
    input_data = opulent_ds.drop_vars(["population", "SF6 (SARGWP100)"]).pr.loc[
        {"animal": ["cow"], "product": ["milk"], "category": ["0", "1"]}
    ]

    # for CH4, we want to exclude the 1 category
    # we want to exclude SF6 fully.
    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source", "scenario (FAOSTAT)"],
        priorities=[
            {"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"},
            {"source": "RAND2021", "scenario (FAOSTAT)": "highpop"},
        ],
        exclude_result=[
            {"entity": "CH4", "category (IPCC 2006)": "1"},
            {"entity": ["SF6", "SF6 (SARGWP100)"]},
        ],
    )
    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            ({"entity": ["CO2", "CH4"]}, primap2.csg.SubstitutionStrategy()),
        ],
    )

    result = primap2.csg.compose(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=strategy_definition,
    )
    # The caller of `compose` is responsible for re-adding priority dimensions
    # if necessary
    result = result.expand_dims(dim={"source": ["composed"]})
    result.pr.ensure_valid()

    assert "CO2" in result.keys()
    assert "Processing of CO2" in result.keys()

    assert_copied_from_input_data(
        result["CO2"],
        input_data["CO2"].loc[{"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"}],
        {},
    )

    assert_copied_from_input_data(
        result["CH4"],
        input_data["CH4"].loc[{"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"}],
        {"category": "0"},
    )

    assert result["CH4"].pr.loc[{"category": "1"}].isnull().all().all()

    result_sf6: xr.DataArray = result["SF6"]
    assert result_sf6.isnull().all().all()


def test_compose_invalid_strategy_definition(opulent_ds):
    """We use an invalid strategy definition and verify it raises an error."""
    input_data = opulent_ds.drop_vars(["population", "SF6 (SARGWP100)"]).pr.loc[
        {
            "animal": ["cow"],
            "product": ["milk"],
            "category": ["0", "1"],
            "area": ["COL"],
        }
    ]
    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source"], priorities=[{"source": "RAND2020"}]
    )
    valid_strategy_definition = primap2.csg.StrategyDefinition(
        [({}, primap2.csg.SubstitutionStrategy())],
    )
    primap2.csg.compose(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=valid_strategy_definition,
    )
    invalid_strategy_definition = primap2.csg.StrategyDefinition(
        [
            ({"notadim": "RAND2020"}, primap2.csg.SubstitutionStrategy()),
            ({}, primap2.csg.SubstitutionStrategy()),
        ],
    )
    with pytest.raises(ValueError, match="is not a valid dimension"):
        primap2.csg.compose(
            input_data=input_data,
            priority_definition=priority_definition,
            strategy_definition=invalid_strategy_definition,
        )


def test_compose_strategy_skipping(opulent_ds):
    """In this test, we use a strategy which raises an error and assert that it is
    skipped properly."""
    input_data = opulent_ds.drop_vars(["population", "SF6 (SARGWP100)"]).pr.loc[
        {"animal": ["cow"], "product": ["milk"], "category": ["0", "1"]}
    ]

    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source", "scenario (FAOSTAT)"],
        priorities=[
            {"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"},
            {"source": "RAND2021", "scenario (FAOSTAT)": "highpop"},
        ],
    )

    # for CH4, we use a strategy which gives up for the RAND2020 source
    class ErroringStrategy:
        type = "erroring"

        def fill(
            self,
            *,
            ts: xr.DataArray,
            fill_ts: xr.DataArray,
            fill_ts_repr: str,
        ) -> tuple[xr.DataArray, list[primap2.ProcessingStepDescription]]:
            raise primap2.csg.StrategyUnableToProcess("no processing")

    # However, we define the substitution strategy as the fallback strategy, so that
    # the Substitution strategy is used for everything anyway.
    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            ({"entity": "CH4", "source": "RAND2020"}, ErroringStrategy()),
            ({}, primap2.csg.SubstitutionStrategy()),
        ]
    )

    result = primap2.csg.compose(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=strategy_definition,
    )
    # The caller of `compose` is responsible for re-adding priority dimensions
    # if necessary
    result = result.expand_dims(dim={"source": ["composed"]})
    result.pr.ensure_valid()

    assert_copied_from_input_data(
        result["CO2"],
        input_data["CO2"].loc[{"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"}],
        {},
    )
    assert_copied_from_input_data(
        result["CH4"],
        input_data["CH4"].loc[{"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"}],
        {},
    )


def test_compose_strategy_all_error(opulent_ds):
    input_data = opulent_ds.drop_vars(["population", "SF6 (SARGWP100)"]).pr.loc[
        {"animal": ["cow"], "product": ["milk"], "category": ["0", "1"]}
    ]

    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source", "scenario (FAOSTAT)"],
        priorities=[
            {"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"},
            {"source": "RAND2021", "scenario (FAOSTAT)": "highpop"},
        ],
    )

    # for CH4, we use a strategy which gives up as the only strategy, which should
    # generate an error
    class ErroringStrategy:
        type = "erroring"

        def fill(
            self,
            *,
            ts: xr.DataArray,
            fill_ts: xr.DataArray,
            fill_ts_repr: str,
        ) -> tuple[xr.DataArray, list[primap2.ProcessingStepDescription]]:
            raise primap2.csg.StrategyUnableToProcess("no processing")

    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            ({"entity": "CH4"}, ErroringStrategy()),
            ({"entity": ["CO2", "SF6"]}, primap2.csg.SubstitutionStrategy()),
        ]
    )

    with pytest.raises(ValueError, match="No configured strategy was able to process"):
        primap2.csg.compose(
            input_data=input_data,
            priority_definition=priority_definition,
            strategy_definition=strategy_definition,
        )


def test_compose_skip_source(opulent_ds):
    """We do not use a specific source for CH4 category 0."""
    input_data = opulent_ds.drop_vars(["population", "SF6 (SARGWP100)", "SF6"]).pr.loc[
        {"animal": ["cow"], "product": ["milk"], "category": ["0", "1"]}
    ]

    # we want to only exclude RAND2020, lowpop, CH4, 0
    # without excluding       RAND2020, lowpop, CO2, 0
    # or                      RAND2020, lowpop, CH4, 1
    # but each dimension is evaluated individually, so we have to select three
    # times, first everything where entity!=CH4 and category!=0, then
    # entity=CH4 but category!=0 and entity!=CH4 but category=0.
    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source", "scenario (FAOSTAT)"],
        priorities=[
            {"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"},
            {"source": "RAND2021", "scenario (FAOSTAT)": "highpop"},
        ],
        exclude_input=[
            {
                "source": "RAND2020",
                "scenario (FAOSTAT)": "lowpop",
                "entity": "CH4",
                "category (IPCC 2006)": "0",
            }
        ],
    )

    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            ({}, primap2.csg.SubstitutionStrategy()),
        ]
    )

    result = primap2.csg.compose(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=strategy_definition,
    )
    # The caller of `compose` is responsible for re-adding priority dimensions
    # if necessary
    result = result.expand_dims(dim={"source": ["composed"]})
    result.pr.ensure_valid()

    assert_copied_from_input_data(
        result["CO2"],
        input_data["CO2"].loc[{"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"}],
        {},
    )
    assert_copied_from_input_data(
        result["CH4"],
        input_data["CH4"].loc[{"source": "RAND2021", "scenario (FAOSTAT)": "highpop"}],
        {"category": "0"},
    )
    assert_copied_from_input_data(
        result["CH4"],
        input_data["CH4"].loc[{"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"}],
        {"category": "1"},
    )

    tpd: primap2.TimeseriesProcessingDescription = (
        result["Processing of CH4"].pr.loc[{"area": "COL", "category": "0"}].item()
    )
    assert len(tpd.steps) == 2
    assert tpd.steps[0].function == "compose_timeseries"
    assert tpd.steps[0].time == "all"
    assert (
        tpd.steps[0].description
        == "{'source': 'RAND2020', 'scenario (FAOSTAT)': 'lowpop'} is excluded from "
        "processing, skipped"
    )
    assert tpd.steps[1].function == "substitution"
    assert tpd.steps[1].source == (
        "{'source': 'RAND2021', 'scenario (FAOSTAT)': 'highpop'}"
    )


def test_compose_skip_variable(opulent_ds):
    """We do skip processing SF6 (SARGWP100) altogether."""
    input_data = opulent_ds.drop_vars(["population"]).pr.loc[
        {
            "animal": ["cow"],
            "product": ["milk"],
            "category": ["0", "1"],
            "area": ["COL"],
        }
    ]

    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source"],
        priorities=[
            {"source": "RAND2020"},
        ],
        exclude_result=[{"variable": "SF6 (SARGWP100)"}],
    )

    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            ({}, primap2.csg.SubstitutionStrategy()),
        ]
    )

    result = primap2.csg.compose(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=strategy_definition,
    )
    # The caller of `compose` is responsible for re-adding priority dimensions
    # if necessary
    result = result.expand_dims(dim={"source": ["composed"]})
    result.pr.ensure_valid()

    assert result["SF6 (SARGWP100)"].isnull().all()
    assert not result["SF6"].isnull().all()


def test_compose_skip_entity(opulent_ds):
    """We do skip processing SF6, even with gwp_context."""
    input_data = opulent_ds.drop_vars(["population"]).pr.loc[
        {
            "animal": ["cow"],
            "product": ["milk"],
            "category": ["0", "1"],
            "area": ["COL"],
        }
    ]

    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source"],
        priorities=[
            {"source": "RAND2020"},
        ],
        exclude_result=[{"entity": "SF6"}],
    )

    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            ({}, primap2.csg.SubstitutionStrategy()),
        ]
    )

    result = primap2.csg.compose(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=strategy_definition,
    )
    # The caller of `compose` is responsible for re-adding priority dimensions
    # if necessary
    result = result.expand_dims(dim={"source": ["composed"]})
    result.pr.ensure_valid()

    assert result["SF6 (SARGWP100)"].isnull().all()
    assert result["SF6"].isnull().all()


def test_compose_variable_entity(opulent_ds):
    """Test that no strategy is found when specifying variable names for entity"""
    input_data = opulent_ds[["SF6 (SARGWP100)"]].pr.loc[
        {
            "animal": ["cow"],
            "product": ["milk"],
            "category": ["0", "1"],
            "area": ["COL"],
        }
    ]

    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source"],
        priorities=[
            {"source": "RAND2020"},
        ],
    )

    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            ({"entity": ["SF6 (SARGWP100)"]}, primap2.csg.SubstitutionStrategy()),
        ]
    )

    with pytest.raises(ValueError, match="No configured strategy was able to process"):
        primap2.csg.compose(
            input_data=input_data,
            priority_definition=priority_definition,
            strategy_definition=strategy_definition,
        )


def test_compose_pbar(opulent_ds):
    input_data = opulent_ds.drop_vars(["population", "SF6 (SARGWP100)"]).pr.loc[
        {"animal": ["cow"], "product": ["milk"], "category": ["0", "1"]}
    ]
    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source", "scenario (FAOSTAT)"],
        priorities=[
            {"source": "RAND2020", "scenario (FAOSTAT)": "lowpop"},
            {"source": "RAND2021", "scenario (FAOSTAT)": "highpop"},
        ],
    )
    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            (
                {},
                primap2.csg.SubstitutionStrategy(),
            )
        ]
    )

    result = primap2.csg.compose(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=strategy_definition,
        progress_bar=None,
    )
    assert "CO2" in result.keys()


def test_compose_sec_cats_missing(opulent_ds):
    """Compose should also work when a dimensions is missing in `sec_cats`."""
    input_data = opulent_ds.drop_vars(["population", "SF6 (SARGWP100)"]).pr.loc[
        {"animal": ["cow"], "category": ["0", "1"]}
    ]
    input_data.attrs["sec_cats"].remove("product (FAOSTAT)")
    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source", "scenario (FAOSTAT)", "product (FAOSTAT)"],
        priorities=[
            {
                "source": "RAND2020",
                "scenario (FAOSTAT)": "lowpop",
                "product (FAOSTAT)": "milk",
            },
            {
                "source": "RAND2021",
                "scenario (FAOSTAT)": "highpop",
                "product (FAOSTAT)": "milk",
            },
        ],
    )
    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            (
                {},
                primap2.csg.SubstitutionStrategy(),
            )
        ]
    )
    primap2.csg.compose(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=strategy_definition,
        progress_bar=None,
    )


def test_compose_timeseries_trivial():
    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source"], priorities=[{"source": "A"}, {"source": "B"}]
    )
    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            ({"source": "B"}, primap2.csg.SubstitutionStrategy()),
            ({"source": "A"}, primap2.csg.SubstitutionStrategy()),
        ]
    )

    time = pd.date_range("1850-01-01", "2022-01-01", freq="YS")
    anp = np.linspace(0.0, 1.0, len(time))
    anp[0] = np.nan
    anp[1] = np.nan
    da_a = get_single_ts(
        coords={"source": "A", "category": "1", "area (ISO3)": "MEX"},
        data=anp,
        time=time,
    )
    bnp = np.linspace(1000.0, 2000.0, len(time))
    bnp[1] = np.nan
    bnp[2] = np.nan
    da_b = get_single_ts(
        coords={"source": "B", "category": "1", "area (ISO3)": "MEX"},
        data=bnp,
        time=time,
    )

    input_data = xr.concat((da_a, da_b), dim="source", join="exact")

    result_ts, result_description = primap2.csg._compose.compose_timeseries(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=strategy_definition,
    )

    enp = np.linspace(0.0, 1.0, len(time))  # same as A
    enp[0] = 1000.0  # filled from B
    enp[1] = np.nan  # not filled, NaN in A and B
    expected_ts = get_single_ts(
        data=enp, coords={"category": "1", "area (ISO3)": "MEX"}, time=time
    )

    xr.testing.assert_identical(result_ts, expected_ts)

    assert len(result_description.steps) == 2
    assert result_description.steps[0].time[0] == np.datetime64("1852-01-01")
    assert result_description.steps[0].function == "substitution"
    assert (
        result_description.steps[0].description
        == "substituted with corresponding values from 'A'"
    )
    assert len(result_description.steps[1].time) == 1
    assert result_description.steps[1].time[0] == np.datetime64("1850-01-01")
    assert result_description.steps[1].function == "substitution"
    assert (
        result_description.steps[1].description
        == "substituted with corresponding values from 'B'"
    )


def test_compose_timeseries_no_match(caplog):
    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source"],
        priorities=[{"source": "C"}, {"source": "A"}, {"source": "B"}],
    )
    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            ({"source": "A"}, primap2.csg.SubstitutionStrategy()),
            ({"source": "B"}, primap2.csg.SubstitutionStrategy()),
        ]
    )
    da_a = get_single_ts(
        coords={"source": "A", "category": "1", "area (ISO3)": "MEX"},
    )
    da_b = get_single_ts(
        coords={"source": "B", "category": "1", "area (ISO3)": "MEX"},
    )
    input_data = xr.concat((da_a, da_b), dim="source", join="exact")

    primap2.csg._compose.compose_timeseries(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=strategy_definition,
    )

    assert "selector={'source': 'C'} matched no input_data, skipping." in caplog.text


def test_compose_timeseries_all_null():
    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source"],
        priorities=[{"source": "A"}, {"source": "B"}],
    )
    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[
            ({"source": "A"}, primap2.csg.SubstitutionStrategy()),
            ({"source": "B"}, primap2.csg.SubstitutionStrategy()),
        ]
    )
    da_a = get_single_ts(
        coords={"source": "A", "category": "1", "area (ISO3)": "MEX"},
    )
    da_a[0] = np.nan
    da_b = get_single_ts(
        data=np.nan,
        coords={"source": "B", "category": "1", "area (ISO3)": "MEX"},
    )
    input_data = xr.concat((da_a, da_b), dim="source", join="exact")

    result, result_description = primap2.csg._compose.compose_timeseries(
        input_data=input_data,
        priority_definition=priority_definition,
        strategy_definition=strategy_definition,
    )

    print(result_description)
    assert result_description.steps[1].description == "'B' is fully NaN, skipped"


def test_compose_timeseries_priorities_wrong():
    priority_definition = primap2.csg.PriorityDefinition(
        priority_dimensions=["source"],
        priorities=[{"source": "C"}],
    )
    strategy_definition = primap2.csg.StrategyDefinition(
        strategies=[({"source": "C"}, primap2.csg.SubstitutionStrategy())]
    )
    da_a = get_single_ts(
        coords={"source": "A", "category": "1", "area (ISO3)": "MEX"},
    )
    da_b = get_single_ts(
        data=np.nan,
        coords={"source": "B", "category": "1", "area (ISO3)": "MEX"},
    )
    input_data = xr.concat((da_a, da_b), dim="source", join="exact")

    with pytest.raises(ValueError):
        primap2.csg._compose.compose_timeseries(
            input_data=input_data,
            priority_definition=priority_definition,
            strategy_definition=strategy_definition,
        )


def test_priority_coordinates_repr():
    assert (
        primap2.csg._compose.priority_coordinates_repr(
            fill_ts=get_single_ts(coords={"source": "A"}),
            priority_dimensions=["source"],
        )
        == "'A'"
    )
    assert (
        primap2.csg._compose.priority_coordinates_repr(
            fill_ts=get_single_ts(coords={"source": "A", "scenario": "S"}),
            priority_dimensions=["source", "scenario"],
        )
        == "{'source': 'A', 'scenario': 'S'}"
    )
    assert (
        primap2.csg._compose.priority_coordinates_repr(
            fill_ts=get_single_ts(coords={"source": "A", "scenario": "S"}),
            priority_dimensions=["scenario"],
        )
        == "'S'"
    )



"""xarray extension accessors providing an API under the 'pr' namespace."""

import xarray as xr

from ._aggregate import DataArrayAggregationAccessor, DatasetAggregationAccessor
from ._data_format import DatasetDataFormatAccessor
from ._downscale import DataArrayDownscalingAccessor, DatasetDownscalingAccessor
from ._fill_combine import DataArrayFillAccessor, DatasetFillAccessor
from ._merge import DataArrayMergeAccessor, DatasetMergeAccessor
from ._metadata import DatasetMetadataAccessor
from ._overview import DataArrayOverviewAccessor, DatasetOverviewAccessor
from ._selection import (
    DataArrayAliasSelectionAccessor,
    DatasetAliasSelectionAccessor,
)
from ._setters import DataArraySettersAccessor, DatasetSettersAccessor
from ._units import DataArrayUnitAccessor, DatasetUnitAccessor


@xr.register_dataset_accessor("pr")
class PRIMAP2DatasetAccessor(
    DatasetAggregationAccessor,
    DatasetAliasSelectionAccessor,
    DatasetDataFormatAccessor,
    DatasetDownscalingAccessor,
    DatasetMergeAccessor,
    DatasetMetadataAccessor,
    DatasetOverviewAccessor,
    DatasetSettersAccessor,
    DatasetUnitAccessor,
    DatasetFillAccessor,
):
    """Collection of methods useful for climate policy analysis."""


@xr.register_dataarray_accessor("pr")
class PRIMAP2DataArrayAccessor(
    DataArrayAggregationAccessor,
    DataArrayAliasSelectionAccessor,
    DataArrayDownscalingAccessor,
    DataArrayMergeAccessor,
    DataArrayOverviewAccessor,
    DataArraySettersAccessor,
    DataArrayUnitAccessor,
    DataArrayFillAccessor,
):
    """Collection of methods useful for climate policy analysis."""



"""Tests for _setters.py"""

import re

import numpy as np
import pint
import pytest
import xarray as xr

from primap2 import ureg

from .utils import assert_aligned_equal, assert_ds_aligned_equal


@pytest.fixture
def da(minimal_ds) -> xr.DataArray:
    da = minimal_ds["CO2"]
    # cast coord explicitly to object, because calling set() often casts to object
    # as a side effect of modifying the coords, and we are fine with that.
    da["area (ISO3)"] = da["area (ISO3)"].astype(object)
    return da


@pytest.fixture
def ts() -> np.ndarray:
    return np.linspace(0, 20, 21)


@pytest.fixture
def co2() -> pint.Unit:
    return ureg("Gg CO2 / year")


@pytest.fixture(params=["fillna_empty", "error", "fillna", "overwrite", None])
def existing(request) -> dict[str, str]:
    if request.param is not None:
        return {"existing": request.param}
    else:
        return {}


@pytest.fixture(params=["error", "extend", None])
def new(request) -> dict[str, str]:
    if request.param is not None:
        return {"new": request.param}
    else:
        return {}


class TestDASetter:
    def test_new_error(self, da: xr.DataArray, ts, co2, existing):
        with pytest.raises(
            KeyError,
            match=re.escape(
                "Values {'CUB'} not in 'area (ISO3)', use new='extend' to automatically"
                " insert new values into dim."
            ),
        ):
            da.pr.set("area", "CUB", ts * co2, new="error", **existing)

    def test_new_works(self, da: xr.DataArray, ts, co2, existing):
        actual = da.pr.set("area", ["CUB"], 2 * ts * co2, new="extend", **existing)
        expected = da.reindex({"area (ISO3)": [*da["area (ISO3)"].values, "CUB"]})
        expected.loc[{"area (ISO3)": "CUB"}] = ts[..., np.newaxis] * 2 * co2
        assert_aligned_equal(actual, expected)

    def test_exists_default_error(
        self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit, new
    ):
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Values {'COL'} for 'area (ISO3)' already exist and contain data."
                " Use existing='overwrite' or 'fillna' to avoid this error."
            ),
        ):
            da.pr.set("area", "COL", ts * co2, **new)

    def test_exists_empty_default(
        self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit, new
    ):
        da.loc[{"area (ISO3)": "COL"}].pint.magnitude[:] = np.nan
        actual = da.pr.set("area", "COL", ts * co2, **new)
        expected = da
        expected.loc[{"area (ISO3)": "COL"}] = ts[..., np.newaxis] * co2
        assert_aligned_equal(actual, expected)

    def test_exists_somena_default(
        self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit, new
    ):
        da.loc[{"area (ISO3)": "COL"}].pint.magnitude[:] = np.nan
        da.loc[{"area (ISO3)": "COL", "time": "2001"}] = 2 * co2
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Values {'COL'} for 'area (ISO3)' already exist and contain data."
                " Use existing='overwrite' or 'fillna' to avoid this error."
            ),
        ):
            da.pr.set("area", "COL", ts * co2, **new)

    def test_exists_error(self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit, new):
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Values {'COL'} for 'area (ISO3)' already exist."
                " Use existing='overwrite' or 'fillna' to avoid this error."
            ),
        ):
            da.pr.set("area", "COL", ts * co2, existing="error", **new)

    def test_exists_overwrite(
        self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit, new
    ):
        actual = da.pr.set("area", "COL", ts * co2, existing="overwrite", **new)
        expected = da.copy()
        expected.loc[{"area (ISO3)": "COL"}] = ts[..., np.newaxis] * co2
        assert_aligned_equal(actual, expected)

    def test_exists_fillna(self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit, new):
        expected = da.copy()
        expected.loc[{"area (ISO3)": "COL", "time": "2009"}] = np.nan * co2
        actual = expected.pr.set("area", "COL", ts * co2, existing="fillna", **new)
        expected.loc[{"area (ISO3)": "COL", "time": "2009"}] = 9 * co2
        assert_aligned_equal(actual, expected)

    def test_mixed_default(self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit):
        da.loc[{"area (ISO3)": "COL"}].pint.magnitude[:] = np.nan
        actual = da.pr.set(
            "area",
            ["COL", "CUB"],
            np.array([ts, 2 * ts]).T * co2,
        )
        expected = da.reindex({"area (ISO3)": [*da["area (ISO3)"].values, "CUB"]})
        expected.loc[{"area (ISO3)": "COL"}] = ts[..., np.newaxis] * co2
        expected.loc[{"area (ISO3)": "CUB"}] = ts[..., np.newaxis] * 2 * co2
        assert_aligned_equal(actual, expected)

    def test_mixed_error(self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit):
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Values {'COL'} for 'area (ISO3)' already exist."
                " Use existing='overwrite' or 'fillna' to avoid this error."
            ),
        ):
            da.pr.set(
                "area",
                ["COL", "CUB"],
                ts * co2,
                existing="error",
            )

    def test_mixed_overwrite(self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit):
        actual = da.pr.set(
            "area",
            ["COL", "CUB"],
            np.array([ts, 2 * ts]).T * co2,
            existing="overwrite",
        )
        expected = da.reindex({"area (ISO3)": [*da["area (ISO3)"].values, "CUB"]})
        expected.loc[{"area (ISO3)": "COL"}] = ts[..., np.newaxis] * co2
        expected.loc[{"area (ISO3)": "CUB"}] = ts[..., np.newaxis] * 2 * co2
        assert_aligned_equal(actual, expected)

        # with explicit value_dims
        actual = da.pr.set(
            "area",
            ["COL", "CUB"],
            np.array([ts, 2 * ts]) * co2,
            value_dims=["area (ISO3)", "time"],
            existing="overwrite",
        )
        assert_aligned_equal(actual, expected)

    def test_mixed_fillna(self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit):
        tda = da.copy()
        tda.loc[{"area (ISO3)": "COL", "time": "2009"}] = np.nan * co2
        actual = tda.pr.set(
            "area",
            ["COL", "CUB"],
            np.array([ts, 2 * ts]).T * co2,
            existing="fillna",
        )
        expected = da.reindex({"area (ISO3)": [*da["area (ISO3)"].values, "CUB"]})
        expected.loc[{"area (ISO3)": "COL", "time": "2009"}] = 9 * co2
        expected.loc[{"area (ISO3)": "CUB"}] = ts[..., np.newaxis] * 2 * co2
        assert_aligned_equal(actual, expected)

    def test_new_from_array(self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit):
        # with scalar dimension
        actual = da.pr.set("area", "CUB", 2 * da.pr.loc[{"area": "COL"}])
        expected = da.reindex({"area (ISO3)": [*da["area (ISO3)"].values, "CUB"]})
        expected = expected.fillna(expected.loc[{"area (ISO3)": "COL"}] * 2)
        assert_aligned_equal(actual, expected)

        actual_error = da.pr.set(
            "area", "CUB", 2 * da.pr.loc[{"area": "COL"}], existing="error"
        )
        assert_aligned_equal(actual_error, expected)

        # with non-scalar dimension
        with pytest.raises(KeyError, match="not all values found in index"):
            da.pr.set("area", "CUB", 2 * da.pr.loc[{"area": ["COL"]}])

    def test_mixed_from_data_array_overwrite(
        self, da: xr.DataArray, ts: np.ndarray, co2: pint.Unit
    ):
        actual = da.pr.set(
            "area", ["CUB", "COL"], 2 * da.pr.loc[{"area": "COL"}], existing="overwrite"
        )
        expected = da.reindex({"area (ISO3)": [*da["area (ISO3)"].values, "CUB"]})
        expected.loc[{"area (ISO3)": "COL"}].pint.magnitude[:] = np.nan
        expected = expected.fillna(da.loc[{"area (ISO3)": "COL"}] * 2)
        assert_aligned_equal(actual, expected)

    @pytest.fixture
    def mda(self):
        return xr.DataArray(
            np.zeros((3, 4, 5)),
            coords=[
                ("a", ["a1", "a2", "a3"]),
                ("b", ["b1", "b2", "b3", "b4"]),
                ("c", ["c1", "c2", "c3", "c4", "c5"]),
            ],
            attrs={"test": "attr"},
            name="thing",
        )

    def test_multidimensional_ndarray(self, mda: xr.DataArray):
        actual = mda.pr.set("a", "a4", np.ones((4, 5)))
        expected = mda.reindex({"a": ["a1", "a2", "a3", "a4"]}).fillna(1)
        assert_aligned_equal(actual, expected)

        actual = mda.pr.set("a", "a4", np.ones(4), value_dims=["b"])
        assert_aligned_equal(actual, expected)

        actual = mda.pr.set("a", "a4", np.ones(5), value_dims=["c"])
        assert_aligned_equal(actual, expected)

        actual = mda.pr.set("a", "a4", np.ones((4, 5)), value_dims=["b", "c"])
        assert_aligned_equal(actual, expected)

    @pytest.mark.parametrize("shape", [3, 4, 5])
    def test_multidimensional_ndarray_underspecified(
        self, mda: xr.DataArray, new, shape
    ):
        match = (
            "Could not automatically determine value dimensions, please use the"
            " value_dims parameter."
        )
        with pytest.raises(ValueError, match=match):
            mda.pr.set("a", "a3", np.ones(shape), existing="overwrite", **new)

    def test_multidimensional_data_array(self, mda: xr.DataArray):
        actual = mda.pr.set(
            "a", "a4", xr.DataArray(np.ones((4, 5)), coords=[mda["b"], mda["c"]])
        )
        expected = mda.reindex({"a": ["a1", "a2", "a3", "a4"]}).fillna(1)
        assert_aligned_equal(actual, expected)

        actual = mda.pr.set("a", "a4", xr.DataArray(np.ones(4), coords=[mda["b"]]))
        assert_aligned_equal(actual, expected)

        actual = mda.pr.set("a", "a4", xr.DataArray(np.ones(5), coords=[mda["c"]]))
        assert_aligned_equal(actual, expected)

    def test_multidimensional_data_array_existing(self, mda: xr.DataArray):
        actual = mda.pr.set(
            "a",
            "a3",
            xr.DataArray(np.ones((4, 5)), coords=[mda["b"], mda["c"]]),
            existing="overwrite",
            new="error",
        )
        expected = mda
        expected.loc[{"a": "a3"}] = 1
        assert_aligned_equal(actual, expected)

        actual = mda.pr.set(
            "a",
            "a3",
            xr.DataArray(np.ones(4), coords=[mda["b"]]),
            existing="overwrite",
            new="error",
        )
        assert_aligned_equal(actual, expected)

        actual = mda.pr.set(
            "a",
            "a3",
            xr.DataArray(np.ones(5), coords=[mda["c"]]),
            existing="overwrite",
            new="error",
        )
        assert_aligned_equal(actual, expected)

    def test_multidimensional_data_array_dim_contained(self, mda: xr.DataArray):
        actual = mda.pr.set(
            "a",
            ["a3", "a4", "a5"],
            xr.DataArray(
                np.ones((3, 4, 5)),
                coords=[("a", ["a3", "a4", "a5"]), mda["b"], mda["c"]],  # type: ignore
            ),
            existing="overwrite",
        )
        expected = mda.reindex({"a": ["a1", "a2", "a3", "a4", "a5"]}).fillna(1)
        expected.loc[{"a": "a3"}] = 1
        assert_aligned_equal(actual, expected)

    def test_multidimensional_data_array_dim_oversized(self, mda: xr.DataArray):
        actual = mda.pr.set(
            "a",
            ["a3", "a4", "a5"],
            xr.DataArray(
                np.ones((4, 4, 5)),
                coords=[
                    ("a", ["a2", "a3", "a4", "a5"]),
                    mda["b"],
                    mda["c"],
                ],  # type: ignore
            ),
            existing="overwrite",
        )
        expected = mda.reindex({"a": ["a1", "a2", "a3", "a4", "a5"]}).fillna(1)
        expected.loc[{"a": "a3"}] = 1
        assert_aligned_equal(actual, expected)

    def test_over_specific(self, da: xr.DataArray, ts: np.ndarray, new):
        with pytest.raises(
            ValueError, match="value_dims given, but value is already a DataArray."
        ):
            da.pr.set(
                "area",
                "COL",
                da.pr.loc[{"area": "BOL"}],
                value_dims=["area", "time", "source"],
                existing="overwrite",
                **new,
            )

    def test_incompatible_units(self, da: xr.DataArray, ts: np.ndarray, new):
        with pytest.raises(pint.errors.DimensionalityError, match="Cannot convert"):
            da.pr.set("area", "COL", ts * ureg("kg"), existing="overwrite", **new)

    def test_automatic_unit_conversion(
        self, da: xr.DataArray, ts: np.ndarray, co2, new
    ):
        actual = da.pr.set(
            "area", "COL", ts * ureg("Mg CO2 / year"), existing="overwrite", **new
        )
        expected = da
        expected.loc[{"area (ISO3)": "COL"}] = 1e-3 * ts[..., np.newaxis] * co2
        assert_aligned_equal(actual, expected)

    def test_dim_does_not_exist(self, da: xr.DataArray, ts, existing, new):
        with pytest.raises(ValueError, match="Dimension 'asdf' does not exist."):
            da.pr.set("asdf", ["COL"], ts, **existing, **new)

    def test_existing_wrong(self, da: xr.DataArray, ts, new):
        with pytest.raises(
            ValueError,
            match="If given, 'existing' must specify one of 'error', 'overwrite', "
            "'fillna_empty', or 'fillna', not 'asdf'.",
        ):
            da.pr.set("area", ["COL"], ts, existing="asdf", **new)

    def test_new_wrong(self, da: xr.DataArray, ts, existing):
        with pytest.raises(
            ValueError,
            match="If given, 'new' must specify one of 'error' or 'extend', not"
            " 'asdf'.",
        ):
            da.pr.set("area", ["CUB"], ts, new="asdf", **existing)


class TestDsSetter:
    def test_new(self, minimal_ds: xr.Dataset, existing):
        actual = minimal_ds.pr.set(
            "area", "CUB", minimal_ds.pr.loc[{"area": "COL"}] * 2, **existing
        )
        expected = minimal_ds.reindex(
            {"area (ISO3)": [*minimal_ds["area (ISO3)"].values, "CUB"]}
        )
        for key in expected.keys():
            expected[key] = expected[key].fillna(
                expected[key].pr.loc[{"area": "COL"}] * 2
            )
        assert_ds_aligned_equal(actual, expected)

    def test_new_error(self, minimal_ds: xr.Dataset, existing):
        with pytest.raises(
            KeyError,
            match=re.escape(
                "Values {'CUB'} not in 'area (ISO3)', use new='extend' to automatically"
                " insert new values into dim."
            ),
        ):
            minimal_ds.pr.set(
                "area",
                "CUB",
                minimal_ds.pr.loc[{"area": "COL"}] * 2,
                new="error",
                **existing,
            )

    def test_existing_default(self, minimal_ds: xr.Dataset, new):
        with pytest.raises(ValueError):
            minimal_ds.pr.set(
                "area", "COL", minimal_ds.pr.loc[{"area": "COL"}] * 2, **new
            )

    def test_existing_overwrite(self, minimal_ds: xr.Dataset, new):
        actual = minimal_ds.pr.set(
            "area",
            "COL",
            minimal_ds.pr.loc[{"area": "COL"}] * 2,
            existing="overwrite",
            **new,
        )
        expected = minimal_ds.pint.dequantify()
        for key in expected:
            expected[key].loc[{"area (ISO3)": "COL"}] = (
                expected[key].loc[{"area (ISO3)": "COL"}] * 2
            )
        expected = expected.pr.quantify()
        assert_ds_aligned_equal(actual, expected)

    def test_existing_fillna(self, minimal_ds: xr.Dataset, new):
        minimal_ds["CO2"].pr.loc[{"area": "COL", "time": "2001"}].pint.magnitude[:] = (
            np.nan
        )
        actual = minimal_ds.pr.set(
            "area", "COL", minimal_ds.pr.loc[{"area": "MEX"}], existing="fillna", **new
        )
        expected = minimal_ds.fillna(minimal_ds.pr.loc[{"area": "MEX"}])
        assert_ds_aligned_equal(actual, expected)

    def test_existing_wrong_type(self, minimal_ds: xr.Dataset, new):
        with pytest.raises(TypeError, match="value must be a Dataset, not"):
            minimal_ds.pr.set("area", "COL", np.zeros((3, 4)), **new)

    def test_wrong_dim(self, minimal_ds: xr.Dataset, existing, new):
        with pytest.raises(ValueError, match="Dimension 'asdf' does not exist."):
            minimal_ds.pr.set(
                "asdf", "COL", minimal_ds.pr.loc[{"area": "COL"}], **existing, **new
            )

    def test_inhomogeneous(self, minimal_ds: xr.Dataset):
        minimal_ds["population"] = minimal_ds["CO2"].pr.dequantify().sum("area (ISO3)")
        actual = minimal_ds.pr.set(
            "area", "CUB", minimal_ds.pr.loc[{"area": "COL"}] * 2
        )
        expected = minimal_ds.reindex(
            {"area (ISO3)": [*minimal_ds["area (ISO3)"].values, "CUB"]}
        )
        for key in expected.keys():
            if key == "population":
                continue
            expected[key] = expected[key].fillna(
                expected[key].pr.loc[{"area": "COL"}] * 2
            )
        assert_ds_aligned_equal(actual, expected)

    def test_opulent_complex(self, opulent_ds: xr.Dataset):
        ds = opulent_ds
        ds.pr.set(
            "category",
            "0",
            ds.pr.loc[{"category": ["1", "2", "3", "4", "5"]}].pr.sum("category"),
            existing="overwrite",
        )



#!/usr/bin/env python
"""Tests for _aggregate.py"""

import pathlib
import re

import numpy as np
import pytest
import xarray as xr

import primap2
from primap2 import ureg

from . import examples
from .utils import assert_equal


@pytest.fixture(params=["opulent_ds", "opulent_ds[CO2]"])
def opulent_ds_or_da(request):
    """Test with the opulent Dataset or an array taken from it."""
    if request.param == "opulent_ds":
        return examples._cached_opulent_ds.copy(deep=True)
    elif request.param == "opulent_ds[CO2]":
        return examples._cached_opulent_ds["CO2"].copy(deep=True)


def test_fill_all_na():
    coords = [("a", [1, 2]), ("b", [1, 2]), ("c", [1, 2, 3])]
    da = xr.DataArray(
        data=[
            [[np.nan, np.nan, np.nan], [np.nan, 1, 2]],
            [[np.nan, np.nan, np.nan], [np.nan, 1, 2]],
        ],
        coords=coords,
    )
    a = da.pr.fill_all_na(dim="b", value=0)
    a_expected = xr.DataArray(
        data=[
            [[0, np.nan, np.nan], [0, 1, 2]],
            [[0, np.nan, np.nan], [0, 1, 2]],
        ],
        coords=coords,
    )
    assert np.allclose(a, a_expected, equal_nan=True)

    b = da.pr.fill_all_na(dim=("a", "b"), value=0)
    assert np.allclose(b, a_expected, equal_nan=True)

    c = da.pr.fill_all_na(dim=("b", "c"), value=0)
    assert np.allclose(c, da, equal_nan=True)

    d = da.pr.fill_all_na(dim="c", value=0)
    d_expected = xr.DataArray(
        data=[
            [[0, 0, 0], [np.nan, 1, 2]],
            [[0, 0, 0], [np.nan, 1, 2]],
        ],
        coords=coords,
    )
    assert np.allclose(d, d_expected, equal_nan=True)

    e = da.pr.fill_all_na(dim=[], value=0)
    assert np.allclose(e, da, equal_nan=True)

    ds = xr.Dataset({"1": da, "2": da.copy()})
    dsf = ds.pr.fill_all_na(dim="b", value=0)
    assert np.allclose(dsf["1"], a_expected, equal_nan=True)
    assert np.allclose(dsf["2"], a_expected, equal_nan=True)


class TestSum:
    def test_skipna_evaluation_dims(self):
        coords = [("a", [1, 2]), ("b", [1, 2]), ("c", [1, 2, 3])]
        da = xr.DataArray(
            data=[
                [[np.nan, np.nan, np.nan], [np.nan, 1, 2]],
                [[np.nan, np.nan, np.nan], [np.nan, 1, 2]],
            ],
            coords=coords,
        )

        b = da.pr.sum(dim="b", skipna_evaluation_dims="a")
        b_expected = xr.DataArray(
            data=[[0, 1, 2], [0, 1, 2]], coords=[coords[0], coords[2]]
        )
        assert np.allclose(b, b_expected, equal_nan=True)

        c = da.pr.sum(dim="b", skipna_evaluation_dims="c")
        c_expected = xr.DataArray(
            data=[[np.nan, 1, 2], [np.nan, 1, 2]], coords=[coords[0], coords[2]]
        )
        assert np.allclose(c, c_expected, equal_nan=True)

        ds = xr.Dataset({"1": da, "2": da.copy()})
        dss = ds.pr.sum(dim="b", skipna_evaluation_dims="a")
        dss_expected = xr.Dataset(
            {
                "1": b_expected,
                "2": b_expected,
            }
        )
        xr.testing.assert_identical(dss, dss_expected)

    def test_skipna(self):
        coords = [("a", [1, 2]), ("b", [1, 2]), ("c", [1, 2, 3])]
        da = xr.DataArray(
            data=[
                [[np.nan, np.nan, np.nan], [np.nan, 1, 2]],
                [[np.nan, np.nan, np.nan], [np.nan, 1, 2]],
            ],
            coords=coords,
        )

        b0 = da.pr.sum(dim="b", skipna=True, min_count=0)
        b0_expected = xr.DataArray(
            data=[[0, 1, 2], [0, 1, 2]], coords=[coords[0], coords[2]]
        )
        assert np.allclose(b0, b0_expected, equal_nan=True)

        b1 = da.pr.sum(dim="b", skipna=True, min_count=1)
        b1_expected = xr.DataArray(
            data=[[np.nan, 1, 2], [np.nan, 1, 2]], coords=[coords[0], coords[2]]
        )
        assert np.allclose(b1, b1_expected, equal_nan=True)

        b2 = da.pr.sum(dim="b", skipna=True, min_count=2)
        b2_expected = xr.DataArray(
            data=[[np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]],
            coords=[coords[0], coords[2]],
        )
        assert np.allclose(b2, b2_expected, equal_nan=True)

        bdef = da.pr.sum(dim="b", skipna=True)
        assert np.allclose(bdef, b1_expected, equal_nan=True)

        ds = xr.Dataset({"1": da, "2": da.copy()})
        dss1 = ds.pr.sum(dim="b", skipna=True, min_count=1)
        dss1_expected = xr.Dataset(
            {
                "1": b1_expected,
                "2": b1_expected,
            }
        )
        xr.testing.assert_identical(dss1, dss1_expected)

        dssdef = ds.pr.sum(dim="b", skipna=True)
        xr.testing.assert_identical(dssdef, dss1_expected)

    def test_inhomogeneous_regression(self, opulent_ds: xr.Dataset):
        ds = opulent_ds
        dss = ds.pr.loc[{"category": ["1", "2", "3", "4", "5"]}].pr.sum("category")
        xr.testing.assert_identical(dss["population"], ds["population"])
        actual = dss["CO2"]
        expected = (
            ds["CO2"]
            .pr.loc[{"category": ["1", "2", "3", "4", "5"]}]
            .sum("category (IPCC 2006)", keep_attrs=True)
        )
        assert_equal(actual, expected)

        # uncomment to refresh this regression test
        # dss.pr.to_netcdf(
        #    pathlib.Path(__file__).parent
        #    / "data"
        #    / "test_sum_skip_allna_inhomogeneous_result.nc"
        # )

        actual = dss
        expected = primap2.open_dataset(
            pathlib.Path(__file__).parent
            / "data"
            / "test_sum_skip_allna_inhomogeneous_result.nc"
        )
        xr.testing.assert_identical(actual, expected)

    def test_errors(self, opulent_ds_or_da):
        with pytest.raises(
            ValueError,
            match="Only one of 'dim' and 'reduce_to_dim' may be supplied, not both.",
        ):
            opulent_ds_or_da.pr.sum("area", reduce_to_dim="category")

        with pytest.raises(
            ValueError,
            match="Only one of 'skipna' and 'skipna_evaluation_dims' may be supplied",
        ):
            opulent_ds_or_da.pr.sum(
                "area", skipna=True, skipna_evaluation_dims="category"
            )

    def test_error_entity(self, opulent_ds):
        with pytest.raises(
            NotImplementedError,
            match="Summing along the entity dimension is only supported",
        ):
            opulent_ds.pr.sum("entity")

    def test_reduce_to_dim_scalar(self, opulent_ds):
        xr.testing.assert_identical(
            opulent_ds[["CO2", "SF6 (SARGWP100)"]].pr.sum(reduce_to_dim="category"),
            opulent_ds[["CO2", "SF6 (SARGWP100)"]].pr.sum(reduce_to_dim=["category"]),
        )

        xr.testing.assert_identical(
            opulent_ds["CO2"].pr.sum(reduce_to_dim="category"),
            opulent_ds["CO2"].pr.sum(reduce_to_dim=["category"]),
        )

    def test_reduce_to_dim(self, opulent_ds: xr.Dataset):
        xr.testing.assert_identical(
            opulent_ds.pr.sum(reduce_to_dim=["entity", "area"]),
            opulent_ds.sum(set(opulent_ds.dims) - {"area (ISO3)"}, keep_attrs=True),
        )

        shared_ds = opulent_ds[["CO2", "SF6 (SARGWP100)"]]
        xr.testing.assert_allclose(
            shared_ds.pr.sum(reduce_to_dim=["area"]),
            shared_ds.sum(set(shared_ds.dims) - {"area (ISO3)"}, keep_attrs=True)
            .to_array("entity")
            .sum("entity", keep_attrs=True),
            atol=0,
        )


def test_any(opulent_ds_or_da):
    xr.testing.assert_identical(
        opulent_ds_or_da.pr.any(dim="area"), opulent_ds_or_da.any(dim="area (ISO3)")
    )


class TestCount:
    def test_inhomogeneous_entity(self, opulent_ds):
        with pytest.raises(NotImplementedError):
            opulent_ds.pr.count("entity")

    def test_entity(self, opulent_ds):
        actual = opulent_ds[["CO2", "SF6"]].pr.count(dim=["entity", "area"])
        assert (actual == 8).all()

    def test_array(self, opulent_ds):
        da = opulent_ds["CO2"]
        actual = da.pr.count(reduce_to_dim=["area"])
        expected = da.count(set(da.dims) - {"area (ISO3)"})
        xr.testing.assert_allclose(actual, expected)

    def test_set(self, opulent_ds):
        ds = opulent_ds
        actual = ds.pr.count(reduce_to_dim=["area", "entity"])
        expected = ds.count(set(ds.dims) - {"area (ISO3)"})
        xr.testing.assert_allclose(actual, expected)


class TestGasBasket:
    # AR4GWP100 values
    sf6 = 22_800
    ch4 = 25

    # AR6GWP100 values
    sf6_ar6 = 25_200
    ch4_ar6 = 27.9

    @pytest.fixture
    def partly_nan_ds(self, empty_ds):
        empty_ds["CO2"][:] = 1 * ureg("Gg CO2 / year")
        empty_ds["SF6"][:] = 1 * ureg("Gg SF6 / year")
        empty_ds["CH4"][:] = 1 * ureg("Gg CH4 / year")
        empty_ds["CH4"].loc[{"area (ISO3)": "COL"}] = np.nan * ureg("Gg CH4 / year")
        return empty_ds

    def test_contents_sum_default(self, partly_nan_ds):
        summed = partly_nan_ds.pr.gas_basket_contents_sum(
            basket="KYOTOGHG (AR4GWP100)",
            basket_contents=["CO2", "SF6", "CH4"],
        )
        expected = partly_nan_ds["KYOTOGHG (AR4GWP100)"].copy()
        expected[:] = (1 + self.sf6 + self.ch4) * ureg("Gg CO2 / year")
        # NaN counted as 0
        expected.loc[{"area (ISO3)": "COL"}] = (1 + self.sf6) * ureg("Gg CO2 / year")
        assert_equal(summed, expected)

    def test_contents_sum_skipna_evaluation_dims(self, partly_nan_ds):
        partly_nan_ds["CH4"].loc[{"area (ISO3)": "ARG", "time": "2012"}] = (
            np.nan * ureg("Gg CH4 / year")
        )
        summed = partly_nan_ds.pr.gas_basket_contents_sum(
            basket="KYOTOGHG (AR4GWP100)",
            basket_contents=["CO2", "SF6", "CH4"],
            skipna_evaluation_dims=("time",),
        )
        expected = partly_nan_ds["KYOTOGHG (AR4GWP100)"].copy()
        expected[:] = (1 + self.sf6 + self.ch4) * ureg("Gg CO2 / year")
        # NaN only skipped where all time points NaN
        expected.loc[{"area (ISO3)": "COL"}] = (1 + self.sf6) * ureg("Gg CO2 / year")
        expected.loc[{"area (ISO3)": "ARG", "time": "2012"}] = np.nan * ureg(
            "Gg CO2 / year"
        )
        assert_equal(summed, expected, equal_nan=True)

    def test_contents_sum_skipna(self, partly_nan_ds):
        summed = partly_nan_ds.pr.gas_basket_contents_sum(
            basket="KYOTOGHG (AR4GWP100)",
            basket_contents=["CO2", "SF6", "CH4"],
            skipna=False,
        )
        expected = partly_nan_ds["KYOTOGHG (AR4GWP100)"].copy()
        expected[:] = (1 + self.sf6 + self.ch4) * ureg("Gg CO2 / year")
        # NaNs not skipped
        expected.loc[{"area (ISO3)": "COL"}] = np.nan * ureg("Gg CO2 / year")
        assert_equal(summed, expected, equal_nan=True)

    def test_contents_sum_new_entity(self, partly_nan_ds):
        summed = partly_nan_ds.pr.gas_basket_contents_sum(
            basket="KYOTOGHG (AR6GWP100)",
            basket_contents=["CO2", "SF6", "CH4"],
        )
        expected = partly_nan_ds["KYOTOGHG (AR4GWP100)"].copy()
        expected[:] = (1 + self.sf6_ar6 + self.ch4_ar6) * ureg("Gg CO2 / year")
        # NaN counted as 0
        expected.loc[{"area (ISO3)": "COL"}] = (1 + self.sf6_ar6) * ureg(
            "Gg CO2 / year"
        )
        expected.name = "KYOTOGHG (AR6GWP100)"
        expected.attrs = {"gwp_context": "AR6GWP100", "entity": "KYOTOGHG"}
        assert_equal(summed, expected)

    @pytest.fixture
    def partly_filled_ds(self, partly_nan_ds):
        partly_nan_ds["KYOTOGHG (AR4GWP100)"][:] = 1 * ureg("Gg CO2 / year")
        partly_nan_ds["KYOTOGHG (AR4GWP100)"].loc[{"area (ISO3)": "COL"}] = (
            np.nan * ureg("Gg CO2 / year")
        )
        partly_nan_ds["KYOTOGHG (AR4GWP100)"].loc[
            {"area (ISO3)": "BOL", "time": "2020"}
        ] = np.nan * ureg("Gg CO2 / year")
        return partly_nan_ds

    def test_fill_na_from_contents_skipna_evaluation_dims(self, partly_filled_ds):
        filled = partly_filled_ds.pr.fill_na_gas_basket_from_contents(
            basket="KYOTOGHG (AR4GWP100)",
            basket_contents=["CO2", "SF6", "CH4"],
            skipna_evaluation_dims=("time",),
        )
        expected = partly_filled_ds["KYOTOGHG (AR4GWP100)"].copy()
        expected.loc[{"area (ISO3)": "COL"}] = (1 + self.sf6) * ureg("Gg CO2 / year")
        expected.loc[{"area (ISO3)": "BOL", "time": "2020"}] = (
            1 + self.sf6 + self.ch4
        ) * ureg("Gg CO2 / year")
        assert_equal(filled, expected)

    def test_fill_na_from_contents_sel(self, partly_filled_ds):
        filled = partly_filled_ds.pr.fill_na_gas_basket_from_contents(
            basket="KYOTOGHG (AR4GWP100)",
            basket_contents=["CO2", "SF6", "CH4"],
            sel={"area (ISO3)": ["BOL"]},
            skipna_evaluation_dims=("time",),
        )
        expected = partly_filled_ds["KYOTOGHG (AR4GWP100)"].copy()
        expected.loc[{"area (ISO3)": "BOL", "time": "2020"}] = (
            1 + self.sf6 + self.ch4
        ) * ureg("Gg CO2 / year")
        assert_equal(filled, expected, equal_nan=True)

        with pytest.raises(
            ValueError,
            match="The dimension of the selection doesn't match the dimension",
        ):
            partly_filled_ds.pr.fill_na_gas_basket_from_contents(
                basket="KYOTOGHG (AR4GWP100)",
                basket_contents=["CO2", "SF6", "CH4"],
                sel={"area (ISO3)": "BOL"},
                skipna_evaluation_dims=("time",),
            )

    def test_fill_na_from_contents_skipna(self, partly_filled_ds):
        filled = partly_filled_ds.pr.fill_na_gas_basket_from_contents(
            basket="KYOTOGHG (AR4GWP100)",
            basket_contents=["CO2", "SF6", "CH4"],
            skipna=False,
        )
        expected = partly_filled_ds["KYOTOGHG (AR4GWP100)"].copy()
        expected.loc[{"area (ISO3)": "COL"}] = np.nan * ureg("Gg CO2 / year")
        expected.loc[{"area (ISO3)": "BOL", "time": "2020"}] = (
            1 + self.sf6 + self.ch4
        ) * ureg("Gg CO2 / year")
        assert_equal(filled, expected, equal_nan=True)

    def test_add_aggregates_variables_no_input(self, partly_filled_ds, caplog):
        """
        test if no input data handled correctly by add_aggregates_variables
        """
        filled = partly_filled_ds.pr.add_aggregates_variables(
            gas_baskets={
                "FGASES (SARGWP100)": ["SF6 (SARGWP100)"],
            },
        )

        assert "FGASES (SARGWP100) not created. No input data present." in caplog.text
        assert "FGASES (SARGWP100)" not in filled.data_vars

    def test_add_aggregates_variables_partial_input(self, partly_filled_ds, caplog):
        """
        test if partial input data handled correctly by add_aggregates_variables
        """
        filled = partly_filled_ds.pr.add_aggregates_variables(
            gas_baskets={
                "FGASES (SARGWP100)": ["SF6", "HFC32"],
            },
        )

        assert (
            "Not all variables present for FGASES (SARGWP100). Missing: " "['HFC32']"
        ) in caplog.text
        assert "FGASES (SARGWP100)" in filled.data_vars
        xr.testing.assert_allclose(
            filled["FGASES (SARGWP100)"],
            filled["SF6"].pr.convert_to_gwp("SARGWP100", "Gg CO2 / a"),
        )
        filled.pr.ensure_valid()

    def test_add_aggregates_variables_tolerance(self, partly_filled_ds, caplog):
        """
        test if the tolerance setting works (the data in partly_filled_ds is
        already inconsistent, so we can use it for the test directly)
        """
        with pytest.raises(
            xr.MergeError,
            match="pr.merge error: found discrepancies " "larger than tolerance",
        ):
            partly_filled_ds.pr.add_aggregates_variables(
                gas_baskets={
                    "KYOTOGHG (AR4GWP100)": ["CO2", "SF6", "CH4"],
                },
            )

        filled = partly_filled_ds.pr.add_aggregates_variables(
            gas_baskets={
                "KYOTOGHG (AR4GWP100)": ["CO2", "SF6", "CH4"],
            },
            tolerance=30000,
        )
        assert "KYOTOGHG (AR4GWP100)" in filled.data_vars
        filled.pr.ensure_valid()

        filled = partly_filled_ds.pr.add_aggregates_variables(
            gas_baskets={
                "KYOTOGHG (AR4GWP100)": {
                    "sources": ["CO2", "SF6", "CH4"],
                    "tolerance": 30000,
                },
            }
        )
        assert "KYOTOGHG (AR4GWP100)" in filled.data_vars

    def test_add_aggregates_variables_filter(self, partly_filled_ds, caplog):
        """
        test if the filter setting works
        """
        reference_ds = partly_filled_ds.pr.add_aggregates_variables(
            gas_baskets={
                "test (SARGWP100)": ["CO2", "SF6", "CH4"],
            },
        )

        filtered_ds = partly_filled_ds.pr.add_aggregates_variables(
            gas_baskets={
                "test (SARGWP100)": {
                    "sources": ["CO2", "SF6", "CH4"],
                    "filter": {"area (ISO3)": ["COL"]},
                },
            },
        )

        assert (
            reference_ds["test (SARGWP100)"].pr.loc[{"area (ISO3)": ["COL"]}]
            == filtered_ds["test (SARGWP100)"].pr.loc[{"area (ISO3)": ["COL"]}]
        ).all()
        assert (
            filtered_ds["test (SARGWP100)"]
            .pr.loc[{"area (ISO3)": ["BOL"]}]
            .isnull()
            .all()
        )

    def test_add_aggregates_variables_error_basket_type(self, partly_filled_ds, caplog):
        """
        test error on unrecognized aggregation definition
        """
        with pytest.raises(
            ValueError,
            match=re.escape("Unrecognized basket type for 'test (SARGWP100)'"),
        ):
            partly_filled_ds.pr.add_aggregates_variables(
                gas_baskets={
                    "test (SARGWP100)": "CO2",
                },
            )


class TestAddAggregatesCoordinates:
    """
    Tests for the add_aggregates_categories method.

    No tests for skipna etc as that is just passed on to sum
    """

    def test_add_aggregates_coordinates_tolerance(self, minimal_ds):
        """
        test if the tolerance setting works
        """

        test_ds = minimal_ds.pr.add_aggregates_coordinates(
            agg_info={
                "area (ISO3)": {
                    "all": {
                        "sources": ["COL", "ARG", "MEX", "BOL"],
                    }
                }
            }
        )

        with pytest.raises(
            xr.MergeError,
            match="pr.merge error: found discrepancies " "larger than tolerance",
        ):
            test_ds.pr.add_aggregates_coordinates(
                agg_info={
                    "area (ISO3)": {
                        "all": {
                            "sources": ["COL", "ARG"],
                        }
                    }
                }
            )

        test_ds.pr.add_aggregates_coordinates(
            agg_info={
                "area (ISO3)": {
                    "all": {"sources": ["COL", "ARG", "MEX", "BOL"], "tolerance": 5}
                }
            }
        )

    def test_add_aggregates_coordinates_result(self, minimal_ds):
        """
        test if aggregated value timeseries are present and correct
        """
        # with the complex configuration
        test_ds = minimal_ds.pr.add_aggregates_coordinates(
            agg_info={
                "area (ISO3)": {
                    "all": {
                        "sources": ["COL", "ARG", "MEX", "BOL"],
                    }
                }
            }
        )

        expected_result = minimal_ds["CO2"].pr.sum(dim="area (ISO3)")
        actual_result = (
            test_ds["CO2"].pr.loc[{"area (ISO3)": ["all"]}].pr.sum(dim="area (ISO3)")
        )
        xr.testing.assert_allclose(expected_result, actual_result)

        # with the simplified configuration
        test_ds = minimal_ds.pr.add_aggregates_coordinates(
            agg_info={
                "area (ISO3)": {
                    "all": ["COL", "ARG", "MEX", "BOL"],
                }
            }
        )

        actual_result = (
            test_ds["CO2"].pr.loc[{"area (ISO3)": ["all"]}].pr.sum(dim="area (ISO3)")
        )
        xr.testing.assert_allclose(expected_result, actual_result)

    def test_add_aggregates_coordinates_result_filter(self, minimal_ds):
        """
        test if filtering works and only the selected time series are actually
        computed and correct
        """
        # filter entity
        test_ds = minimal_ds.pr.add_aggregates_coordinates(
            agg_info={
                "area (ISO3)": {
                    "all": {
                        "sources": ["COL", "ARG", "MEX", "BOL"],
                        "filter": {"entity": ["SF6"]},
                    }
                }
            }
        )

        # as we filter for entity we expect the variables SF6 and SF6 (SARGWP100)
        # to be aggregated but for CO2 we expect np.nan
        expected_result_SF6 = minimal_ds["SF6"].pr.sum(dim="area (ISO3)")
        actual_result_SF6 = (
            test_ds["SF6"].pr.loc[{"area (ISO3)": ["all"]}].pr.sum(dim="area (ISO3)")
        )
        xr.testing.assert_allclose(expected_result_SF6, actual_result_SF6)
        expected_result_SF6GWP = minimal_ds["SF6 (SARGWP100)"].pr.sum(dim="area (ISO3)")
        actual_result_SF6GWP = (
            test_ds["SF6 (SARGWP100)"]
            .pr.loc[{"area (ISO3)": ["all"]}]
            .pr.sum(dim="area (ISO3)")
        )
        xr.testing.assert_allclose(expected_result_SF6GWP, actual_result_SF6GWP)

        expected_result_CO2 = xr.full_like(expected_result_SF6, np.nan).pr.quantify(
            units="Gg CO2 / year"
        )
        actual_result_CO2 = (
            test_ds["CO2"]
            .pr.loc[{"area (ISO3)": ["all"]}]
            .pr.sum(dim="area (ISO3)", skipna=True, min_count=1)
        )
        xr.testing.assert_allclose(expected_result_CO2, actual_result_CO2)

        # filter variable
        test_ds = minimal_ds.pr.add_aggregates_coordinates(
            agg_info={
                "area (ISO3)": {
                    "all": {
                        "sources": ["COL", "ARG", "MEX", "BOL"],
                        "filter": {"variable": ["SF6"]},
                    }
                }
            }
        )

        # as we filter for entity we expect the variables SF6 and SF6 (SARGWP100)
        # to be aggregated but for CO2 we expect np.nan
        expected_result_SF6 = minimal_ds["SF6"].pr.sum(dim="area (ISO3)")
        actual_result_SF6 = (
            test_ds["SF6"].pr.loc[{"area (ISO3)": ["all"]}].pr.sum(dim="area (ISO3)")
        )
        xr.testing.assert_allclose(expected_result_SF6, actual_result_SF6)

        expected_result_SF6GWP = xr.full_like(expected_result_SF6, np.nan).pr.quantify(
            units="Gg CO2 / year"
        )
        # expected_result_SF6GWP = expected_result_SF6GWP.pr.convert_to_gwp
        actual_result_SF6GWP = (
            test_ds["SF6 (SARGWP100)"]
            .pr.loc[{"area (ISO3)": ["all"]}]
            .pr.sum(dim="area (ISO3)", skipna=True, min_count=1)
        )
        xr.testing.assert_allclose(expected_result_SF6GWP, actual_result_SF6GWP)

        expected_result_CO2 = xr.full_like(expected_result_SF6, np.nan).pr.quantify(
            units="Gg CO2 / year"
        )
        actual_result_CO2 = (
            test_ds["CO2"]
            .pr.loc[{"area (ISO3)": ["all"]}]
            .pr.sum(dim="area (ISO3)", skipna=True, min_count=1)
        )
        xr.testing.assert_allclose(expected_result_CO2, actual_result_CO2)

    def test_add_aggregates_coordinates_warning(self, minimal_ds, caplog):
        """
        Test warnings
        """
        # test warning for missing input
        minimal_ds.pr.add_aggregates_coordinates(
            agg_info={
                "area": {
                    "all": {
                        "sources": ["COL", "ARG", "MEX", "BOL", "DEU"],
                    }
                }
            }
        )
        assert (
            "Not all source values present for 'all'"
            " in coordinate 'area (ISO3)'. "
            "Missing: {'DEU'}"
        ) in caplog.text

        # test warning for fully missing input
        minimal_ds.pr.add_aggregates_coordinates(
            agg_info={
                "area (ISO3)": {
                    "all": {
                        "sources": ["DEU"],
                    }
                }
            }
        )
        assert (
            "No source value present for 'all' in "
            "coordinate 'area (ISO3)'. Missing: {'DEU'}."
        ) in caplog.text

        # test all nan warning
        test_ds = minimal_ds.copy(deep=True)
        test_ds["CO2"] = xr.full_like(test_ds["CO2"], np.nan).pr.quantify(
            units="Gg CO2 / year"
        )
        test_ds.pr.add_aggregates_coordinates(
            agg_info={
                "area (ISO3)": {
                    "all": {
                        "sources": ["COL", "ARG", "MEX", "BOL"],
                    }
                }
            }
        )
        assert (
            "All input data nan for 'all' in " "coordinate 'area (ISO3)'."
        ) in caplog.text

    def test_add_aggregates_coordinates_errors(self, minimal_ds):
        """
        test errors
        """
        # additional coordinate not present
        with pytest.raises(
            ValueError,
            match="Additional coordinate 'area_name' specified but not "
            "present in data",
        ):
            minimal_ds.pr.add_aggregates_coordinates(
                agg_info={
                    "area (ISO3)": {
                        "all": {
                            "sources": ["COL", "ARG", "MEX", "BOL"],
                            "area_name": "All Countries",
                        }
                    }
                }
            )

        # unrecognized aggregation definition
        with pytest.raises(
            ValueError, match="Unrecognized aggregation definition for 'all'"
        ):
            minimal_ds.pr.add_aggregates_coordinates(
                agg_info={"area (ISO3)": {"all": "COL"}}
            )

    def test_add_aggregates_coordinates_add_coord(self, minimal_ds):
        """test error on additional coordinate"""
        test_ds = minimal_ds.copy(deep=True)
        area_name = ["COL", "ARG", "MEX", "BOL"]
        test_ds = test_ds.assign_coords(area_name=("area (ISO3)", area_name))

        test_ds = test_ds.pr.add_aggregates_coordinates(
            agg_info={
                "area (ISO3)": {
                    "all": {
                        "sources": ["COL", "ARG", "MEX", "BOL"],
                        "area_name": "All Countries",
                    }
                }
            }
        )

        assert "All Countries" in test_ds.coords["area_name"]


# filter, also for individual items in the list and check if results fine
# filter for entity, variable and a coordinate



"""workaround for xarray coordinate bug

Additional coordinates are lost during operations that require aligning if the
coordinates differ for the structures to be aligned.

see https://github.com/pydata/xarray/issues/9124

"""

import xarray as xr
from loguru import logger

from ._accessor_base import BaseDataArrayAccessor, BaseDatasetAccessor


class DataArrayFillAccessor(BaseDataArrayAccessor):
    def fillna(self: xr.DataArray, da_fill: xr.DataArray) -> xr.DataArray:
        """
        Wrapper for `fillna` which ensures that additional coordinates present in
        the calling DataArray are also present in the result. The default xarray
        `fillna` implementation silently drops additional (non-indexed) coordinates
        if they need alignment (do not cover the same values in both DataArrays)


        Parameters
        ----------
        da_fill
            DataArray used to fill nan values in the calling DataArray

        Returns
        -------
            calling DatArray where nan values are filled from da_fill where possible

        """
        da_start = self._da

        coords_start = da_start.coords
        coords_fill = da_fill.coords
        filled = da_start.fillna(da_fill)
        # check if we lost a coordinate
        missing_coords = set(coords_start).difference(set(filled.coords))
        for coord in missing_coords:
            try:
                # merge and align the coordinate
                logger.info(f"adding coord: {coord}")
                coords_merged = coords_start[coord].fillna(coords_fill[coord])
                aligned = xr.align(filled, coords_merged, join="outer")
                # set the coordinate
                filled = aligned[0].assign_coords(
                    {coord: (coords_start[coord].dims, aligned[1].data)}
                )
            except Exception as ex:
                message = f"Could not re-add lost coordinate {coord}: {ex}"
                logger.error(message)
                raise ValueError(message) from ex

        return filled

    def combine_first(self: xr.DataArray, da_combine: xr.DataArray) -> xr.DataArray:
        """
        Wrapper for `combine_first` which ensures that additional coordinates
        present in the calling dataset are also present in the result. The default
        xarray `fillna` implementation silently drops additional (non-indexed)
        coordinates if they need alignment (do not cover the same values in both
        DataArrays)


        Parameters
        ----------
        da_combine
            DataArray used to combine with the calling DataArray

        Returns
        -------
            calling DataArray combined with da_combine

        """
        da_start = self._da

        coords_start = da_start.coords
        coords_fill = da_combine.coords
        filled = da_start.combine_first(da_combine)
        # check if we lost a coordinate
        missing_coords = set(coords_start).difference(set(filled.coords))
        for coord in missing_coords:
            try:
                # merge and align the coordinate
                logger.info(f"adding coord: {coord}")
                coords_merged = coords_start[coord].combine_first(coords_fill[coord])
                aligned = xr.align(filled, coords_merged, join="outer")
                # set the coordinate
                filled = aligned[0].assign_coords(
                    {coord: (coords_start[coord].dims, aligned[1].data)}
                )
            except Exception as ex:
                message = f"Could not re-add lost coordinate {coord}: {ex}"
                logger.error(message)
                raise ValueError(message) from ex

        return filled


class DatasetFillAccessor(BaseDatasetAccessor):
    def fillna(
        self: xr.Dataset,
        ds_fill: xr.Dataset | xr.DataArray,
    ) -> xr.Dataset:
        """
        Wrapper for `fillna` which ensures that additional coordinates present in
        the calling Dataset are also present in the result. The default xarray
        `fillna` implementation silently drops additional (non-indexed) coordinates
        if they need alignment (do not cover the same values in both Datasets)


        Parameters
        ----------
        ds_fill
            Dataset or DataArray used to fill nan values in the calling dataset

        Returns
        -------
            calling Dataset where nan values are filled from ds_fill where possible

        """

        if self._ds.pr.has_processing_info():
            raise NotImplementedError(
                "Dataset contains processing information, this is not supported yet. "
                "Use ds.pr.remove_processing_info()."
            )

        ds_start = self._ds

        coords_start = ds_start.coords
        coords_fill = ds_fill.coords
        filled = ds_start.fillna(ds_fill)
        # check if we lost a coordinate
        missing_coords = set(coords_start).difference(set(filled.coords))
        for coord in missing_coords:
            try:
                # merge and align the coordinate
                logger.info(f"adding coord: {coord}")
                coords_merged = coords_start[coord].fillna(coords_fill[coord])
                aligned = xr.align(filled, coords_merged, join="outer")
                # set the coordinate
                filled = aligned[0].assign_coords(
                    {coord: (coords_start[coord].dims, aligned[1].data)}
                )
            except Exception as ex:
                message = f"Could not re-add lost coordinate {coord}: {ex}"
                logger.error(message)
                raise ValueError(message) from ex

        return filled

    def combine_first(
        self: xr.Dataset,
        ds_combine: xr.Dataset | xr.DataArray,
    ) -> xr.Dataset:
        """
        Wrapper for `fillna` which ensures that additional coordinates present in
        the calling Dataset are also present in the result. The default xarray
        `fillna` implementation silently drops additional (non-indexed) coordinates
        if they need alignment (do not cover the same values in both Datasets)


        Parameters
        ----------
        ds_combine
            Dataset or DataArray to combine with the calling Dataset

        Returns
        -------
            calling Dataset calling DataArray combined with da_combine

        """

        if self._ds.pr.has_processing_info():
            raise NotImplementedError(
                "Dataset contains processing information, this is not supported yet. "
                "Use ds.pr.remove_processing_info()."
            )

        ds_start = self._ds

        coords_start = ds_start.coords
        coords_fill = ds_combine.coords
        filled = ds_start.combine_first(ds_combine)
        # check if we lost a coordinate
        missing_coords = set(coords_start).difference(set(filled.coords))
        for coord in missing_coords:
            try:
                # merge and align the coordinate
                logger.info(f"adding coord: {coord}")
                coords_merged = coords_start[coord].combine_first(coords_fill[coord])
                aligned = xr.align(filled, coords_merged, join="outer")
                # set the coordinate
                filled = aligned[0].assign_coords(
                    {coord: (coords_start[coord].dims, aligned[1].data)}
                )
            except Exception as ex:
                message = f"Could not re-add lost coordinate {coord}: {ex}"
                logger.error(message)
                raise ValueError(message) from ex

        return filled



from collections.abc import Hashable, Sequence

import xarray as xr

from ._accessor_base import BaseDataArrayAccessor, BaseDatasetAccessor
from ._aggregate import select_no_scalar_dimension
from ._units import ureg


class DataArrayDownscalingAccessor(BaseDataArrayAccessor):
    def downscale_timeseries(
        self,
        *,
        dim: Hashable,
        basket: Hashable,
        basket_contents: Sequence[Hashable],
        check_consistency: bool = True,
        sel: dict[Hashable, Sequence] | None = None,
        skipna_evaluation_dims: None | Sequence[Hashable] = None,
        skipna: bool = True,
        tolerance: float = 0.01,
    ) -> xr.DataArray:
        """Downscale timeseries along a dimension using a basket defined on a
        broader timeseries.

        This is useful if you have data for many points in time for a total, for example
        the entire Energy sector, and higher-resolution data (e.g. fossil and non-fossil
        energies separately) for only a few points in time. In the example, the Energy
        sector is the ``basket`` and fossil and non-fossil energies are the basket
        contents.
        From any time points where all the basket contents are known, the
        relative shares of the basket contents are determined, and then interpolated
        linearly and extrapolated constantly to the full timeseries. The shares are then
        used to downscale the basket to its contents, which is used to fill gaps in the
        timeseries of the basket contents.

        Parameters
        ----------
        dim: str
          The name of the dimension which contains the basket and its contents, has to
          be one of the dimensions in ``ds.dims``.
        basket: str
          The name of the super-category for which values are known at higher temporal
          resolution and/or for a wider range. A value from ``ds[dimension]``.
        basket_contents: list of str
          The name of the sub-categories. The sum of all sub-categories equals the
          basket. Values from ``ds[dimension]``.
        check_consistency: bool, default True
          If for all points where the basket and all basket_contents are defined,
          it should be checked if the sum of the basket_contents actually equals
          the basket. A ``ValueError`` is raised if the consistency check fails.
        sel: Selection dict, optional
          If the downscaling should only be done on a subset of the Dataset while
          retaining all other values unchanged, give a selection dictionary. The
          downscaling will be done on ``ds.loc[sel]``.
        skipna_evaluation_dims: list of str, optional
          Dimensions which should be evaluated to determine if NA values should be
          skipped entirely if missing fully. By default, no NA values are skipped.
        skipna: bool, default True
          If true it will be passed on to xarray's ds.sum function with min_count=1
          for the calculation of the basket.
          The effect is that NA values in a sum will be ignored and treated as zero
          in the sum unless all values are NA which results in NA.
        tolerance: float
          If given it overrides the default tolerance for deviations of sums of
          individual timeseries to given aggregate timeseries. Default is 0.01 (1%)

        Returns
        -------
        downscaled: xr.DataArray
        """
        da_sel = select_no_scalar_dimension(self._da, sel)

        basket_contents_da = da_sel.loc[{dim: basket_contents}]
        basket_da = da_sel.loc[{dim: basket}]

        if skipna_evaluation_dims is not None:
            if skipna:
                raise ValueError(
                    "Only one of 'skipna' and 'skipna_evaluation_dims' may be supplied, not"
                    " both."
                )
            else:
                skipna = None

        basket_sum = basket_contents_da.pr.sum(
            dim=dim,
            skipna=skipna,
            min_count=1,
            skipna_evaluation_dims=skipna_evaluation_dims,
        )

        if check_consistency:
            deviation: xr.DataArray = abs(basket_da / basket_sum - 1)
            devmax = float(deviation.max().pint.dequantify().data)
            if devmax > tolerance:
                raise ValueError(
                    f"Sum of the basket_contents {basket_contents!r} deviates"
                    f" {devmax * 100} % from the basket"
                    f" {basket!r}, which is more than the allowed {tolerance*100}%. "
                    "To continue regardless, set check_consistency=False."
                )

        # inter- and extrapolate
        shares: xr.DataArray = (
            (basket_contents_da / basket_sum)
            .pint.to("")
            .pint.dequantify()
            .interpolate_na(dim="time", method="linear")
            .ffill(dim="time")
            .bfill(dim="time")
        )

        downscaled: xr.DataArray = basket_da * shares

        return self._da.fillna(downscaled)


class DatasetDownscalingAccessor(BaseDatasetAccessor):
    def downscale_timeseries(
        self,
        *,
        dim: Hashable,
        basket: Hashable,
        basket_contents: Sequence[Hashable],
        check_consistency: bool = True,
        sel: dict[Hashable, Sequence] | None = None,
        skipna_evaluation_dims: Sequence[Hashable] | None = None,
        skipna: bool = True,
        tolerance: float = 0.01,
    ) -> xr.Dataset:
        """Downscale timeseries along a dimension using a basket defined on a
        broader timeseries.

        This is useful if you have data for many points in time for a total, for example
        the entire Energy sector, and higher-resolution data (e.g. fossil and non-fossil
        energies separately) for only a few points in time. In the example, the Energy
        sector is the ``basket`` and fossil and non-fossil energies are the basket
        contents.
        From any time points where all the basket contents are known, the
        relative shares of the basket contents are determined, and then interpolated
        linearly and extrapolated constantly to the full timeseries. The shares are then
        used to downscale the basket to its contents, which is used to fill gaps in the
        timeseries of the basket contents.

        Parameters
        ----------
        dim: str
          The name of the dimension which contains the basket and its contents, has to
          be one of the dimensions in ``ds.dims``.
        basket: str
          The name of the super-category for which values are known at higher temporal
          resolution and/or for a wider range. A value from ``ds[dimension]``.
        basket_contents: list of str
          The name of the sub-categories. The sum of all sub-categories equals the
          basket. Values from ``ds[dimension]``.
        check_consistency: bool, default True
          If for all points where the basket and all basket_contents are defined,
          it should be checked if the sum of the basket_contents actually equals
          the basket. A ``ValueError`` is raised if the consistency check fails.
        sel: Selection dict, optional
          If the downscaling should only be done on a subset of the Dataset while
          retaining all other values unchanged, give a selection dictionary. The
          downscaling will be done on ``ds.loc[sel]``.
        skipna_evaluation_dims: list of str, optional
          Dimensions which should be evaluated to determine if NA values should be
          skipped entirely if missing fully. By default, no NA values are skipped.
        skipna: bool, default True
          If true it will be passed on to xarray's ds.sum function with min_count=1 for
          the calculation of the basket.
          The effect is that NA values in a sum will be ignored and treated as zero
          in the sum unless all values are NA which results in NA.
        tolerance: float
          If given it overrides the default tolerance for deviations of sums of
          individual timeseries to given aggregate timeseries. Default is 0.01 (1%)

        Notes
        -----
        To downscale along the entity dimension, i.e. to downscale gases, see
        :py:meth:`downscale_gas_timeseries`, which handles gwp conversions
        appropriately.

        Returns
        -------
        downscaled: xr.Dataset
        """
        if self._ds.pr.has_processing_info():
            raise NotImplementedError(
                "Dataset contains processing information, this is not supported yet. "
                "Use ds.pr.remove_processing_info()."
            )

        ds_sel = select_no_scalar_dimension(self._ds, sel)

        basket_contents_ds = ds_sel.loc[{dim: basket_contents}]
        basket_ds = ds_sel.loc[{dim: basket}]

        if skipna_evaluation_dims is not None:
            if skipna:
                raise ValueError(
                    "Only one of 'skipna' and 'skipna_evaluation_dims' may be supplied, not"
                    " both."
                )
            else:
                skipna = None

        basket_sum = basket_contents_ds.pr.sum(
            dim=dim,
            skipna=skipna,
            min_count=1,
            skipna_evaluation_dims=skipna_evaluation_dims,
        )

        if check_consistency:
            deviation = abs(basket_ds / basket_sum - 1)
            devmax = deviation.to_array().max()
            if devmax > tolerance:
                raise ValueError(
                    f"Sum of the basket_contents {basket_contents!r} deviates"
                    f" {devmax * 100} % from the basket"
                    f" {basket!r}, which is more than the allowed 1 %. "
                    "To continue regardless, set check_consistency=False."
                )

        # inter- and extrapolate
        shares: xr.Dataset = (
            (basket_contents_ds / basket_sum)
            .pint.to({x: "" for x in basket_contents_ds.keys()})
            .pint.dequantify()
            .interpolate_na(dim="time", method="linear")
            .ffill(dim="time")
            .bfill(dim="time")
        )

        downscaled: xr.Dataset = basket_ds * shares

        return self._ds.fillna(downscaled)

    def downscale_gas_timeseries(
        self,
        *,
        basket: Hashable,
        basket_contents: Sequence[Hashable],
        check_consistency: bool = True,
        sel: dict[Hashable, Sequence] | None = None,
        skipna_evaluation_dims: Sequence[Hashable] | None = None,
        skipna: bool = True,
        tolerance: float = 0.01,
    ) -> xr.Dataset:
        """Downscale a gas basket defined on a broader timeseries to its contents
        known on fewer time points.

        This is useful if you have data for many points in time for a gas basket, for
        example KYOTOGHG, and higher-resolution data (e.g. the individual green house
        gases included in KYOTOGHG) for only a few points in time.
        From any time points where all the basket contents are known, the relative
        shares of the basket contents are determined, and then interpolated linearly
        and extrapolated constantly to the full timeseries. The shares are then
        used to downscale the basket to its contents, which is used to fill gaps in the
        timeseries of the basket contents.
        Basket contents are converted to the units of the basket for downscaling, and
        converted back afterwards; for both conversions the gwp conversions of the
        basket are used.

        Parameters
        ----------
        basket: str
          The name of the gas basket for which values are known at higher temporal
          resolution and/or for a wider range. A value from ``ds.keys()``.
        basket_contents: list of str
          The name of the gases in the gas basket. The sum of all basket_contents
          equals the basket. Values from ``ds.keys()``.
        check_consistency: bool, default True
          If for all points where the basket and all basket_contents are defined,
          it should be checked if the sum of the basket_contents actually equals
          the basket. A ``ValueError`` is raised if the consistency check fails.
        sel: Selection dict, optional
          If the downscaling should only be done on a subset of the Dataset while
          retaining all other values unchanged, give a selection dictionary. The
          downscaling will be done on ``ds.loc[sel]``.
        skipna_evaluation_dims: list of str, optional
          Dimensions which should be evaluated to determine if NA values should be
          skipped entirely if missing fully. By default, no NA values are skipped.
        skipna: bool, optional
          If true it will be passed on to xarray's ds.sum function with min_count=1 for
          the calculation of the basket.
          The effect is that NA values in a sum will be ignored and treated as zero
          in the sum unless all values are NA which results in NA.
        tolerance: float
          If given it overrides the default tolerance for deviations of sums of
          individual timeseries to given aggregate timeseries. Default is 0.01 (1%)

        Returns
        -------
        downscaled: xr.Dataset
        """
        if self._ds.pr.has_processing_info():
            raise NotImplementedError(
                "Dataset contains processing information, this is not supported yet. "
                "Use ds.pr.remove_processing_info()."
            )

        ds_sel = select_no_scalar_dimension(self._ds, sel)

        basket_contents_converted = xr.Dataset()
        da_basket = ds_sel[basket]
        for var in basket_contents:
            da: xr.DataArray = ds_sel[var]
            basket_contents_converted[var] = da.pr.convert_to_gwp_like(like=da_basket)

        if skipna_evaluation_dims is not None:
            if skipna:
                raise ValueError(
                    "Only one of 'skipna' and 'skipna_evaluation_dims' may be supplied, not"
                    " both."
                )
            else:
                skipna = None

        basket_sum = basket_contents_converted.pr.sum(
            dim="entity",
            skipna=skipna,
            min_count=1,
            skipna_evaluation_dims=skipna_evaluation_dims,
        )

        if check_consistency:
            deviation = abs(da_basket / basket_sum - 1)
            devmax = deviation.max().item()
            if devmax > tolerance:
                raise ValueError(
                    f"Sum of the basket_contents {basket_contents!r} deviates"
                    f" {devmax * 100} % from the basket"
                    f" {basket!r}, which is more than the allowed "
                    f"{tolerance * 100}%. "
                    f" {deviation}"
                    "To continue regardless, set check_consistency=False."
                )

        # inter- and extrapolate
        shares = (
            (basket_contents_converted / basket_sum)
            .pint.to({x: "" for x in basket_contents_converted.keys()})
            .pint.dequantify()
            .interpolate_na(dim="time", method="linear")
            .ffill(dim="time")
            .bfill(dim="time")
        )

        downscaled = da_basket * shares

        downscaled_converted = xr.Dataset()

        with ureg.context(da_basket.attrs["gwp_context"]):
            for var in basket_contents:
                downscaled_converted[var] = ds_sel[var].fillna(
                    downscaled[var].pint.to(ds_sel[var].pint.units)
                )

        return self._ds.pr.fillna(downscaled_converted)



"""Simple strategy which replaces NaNs by datapoints from second timeseries."""

import attrs
import xarray as xr

import primap2


@attrs.define(frozen=True)
class SubstitutionStrategy:
    """Fill missing data in the result dataset by copying.

    The NaNs in the result dataset are substituted with data from the filling
    dataset.
    """

    type = "substitution"

    def fill(
        self,
        *,
        ts: xr.DataArray,
        fill_ts: xr.DataArray,
        fill_ts_repr: str,
    ) -> tuple[xr.DataArray, list[primap2.ProcessingStepDescription]]:
        """Fill gaps in ts using data from the fill_ts.

        Parameters
        ----------
        ts
            Base timeseries. Missing data (NaNs) in this timeseries will be filled.
            This function does not modify the data in ts.
        fill_ts
            Fill timeseries. Data from this timeseries will be used (possibly after
            modification) to fill missing data in the base timeseries.
            This function does not modify the data in fill_ts.
        fill_ts_repr
            String representation of fill_ts. Human-readable short representation of
            the fill_ts (e.g. the source).

        Returns
        -------
            filled_ts, descriptions. filled_ts contains the result, where missing
            data in ts is (partly) filled using unmodified data from fill_ts.
            descriptions contains information about which years were affected and
            filled how.
        """
        filled_ts = xr.core.ops.fillna(ts, fill_ts, join="exact")
        filled_mask = ts.isnull() & ~fill_ts.isnull()
        time_filled = (
            "all" if filled_mask.all() else filled_mask["time"][filled_mask].to_numpy()
        )
        description = primap2.ProcessingStepDescription(
            time=time_filled,
            description="substituted with corresponding values from" f" {fill_ts_repr}",
            function=self.type,
            source=fill_ts_repr,
        )
        return filled_ts, [description]



import typing

import xarray as xr

DatasetOrDataArray = typing.TypeVar("DatasetOrDataArray", xr.Dataset, xr.DataArray)
KeyT = typing.TypeVar("KeyT", str, typing.Mapping[typing.Hashable, typing.Any])
DimOrDimsT = typing.TypeVar(
    "DimOrDimsT",
    str,
    typing.Hashable,
    typing.Iterable[str],
    typing.Iterable[typing.Hashable],
)
FunctionT = typing.TypeVar("FunctionT", bound=typing.Callable[..., typing.Any])



import typing

import numpy as np
import xarray as xr

from . import _accessor_base
from ._selection import alias_dims


class DataArraySettersAccessor(_accessor_base.BaseDataArrayAccessor):
    @staticmethod
    def _sel_error(
        da: xr.DataArray, dim: typing.Hashable, key: typing.Iterable
    ) -> xr.DataArray:
        try:
            return da.loc[{dim: key}]
        except KeyError:
            missing = set(key).difference(set(da[dim].values))
            raise KeyError(
                f"Values {missing!r} not in {dim!r}, use new='extend' to "
                f"automatically insert new values into dim."
            ) from None

    @alias_dims(["dim", "value_dims"])
    def set(
        self,
        dim: typing.Hashable,
        key: typing.Any,
        value: xr.DataArray | np.ndarray,
        *,
        value_dims: list[typing.Hashable] | None = None,
        existing: str = "fillna_empty",
        new: str = "extend",
    ) -> xr.DataArray:
        """Set values, optionally expanding the given dimension as necessary.

        The handling of already existing key values can be selected using the
        ``existing`` parameter.

        Parameters
        ----------
        dim: str
            Dimension along which values should be set.
        key: scalar or list of scalars
            Keys in the dimension which should be set. Key values which are missing
            in the dimension are inserted. The handling of key values which already
            exist in the dimension is determined by the ``existing`` parameter.
        value: xr.DataArray or np.ndarray
            Values that will be inserted at the positions specified by ``key``.
            ``value`` needs to be broadcastable to ``da[{dim: key}]``.
        value_dims: list of str, optional
            Specifies the dimensions of ``value``. If ``value`` is not a DataArray
            and ``da[{dim: key}]`` is higher-dimensional, it is necessary to specify
            the value dimensions.
        existing: "fillna_empty", "error", "overwrite", or "fillna", optional
            How to handle existing keys. If ``existing="fillna_empty"`` (default), new
            values overwrite existing values only if all existing values are NaN.
            If ``existing="error"``, a ValueError is raised if any key already exists
            in the index. If ``existing="overwrite"``, new values overwrite current
            values for existing keys. If ``existing="fillna"``, the new values only
            overwrite NaN values for existing keys.
        new: "extend", or "error", optional
            How to handle new keys. If ``new="extend"`` (default), keys which do not
            exist so far are automatically inserted by extending the dimension.
            If ``new="error"``, a KeyError is raised if any key is not yet in the
            dimension.

        Examples
        --------
        >>> import pandas as pd
        >>> import xarray as xr
        >>> import numpy as np
        >>> da = xr.DataArray(
        ...     [[0.0, 1.0, 2.0, 3.0], [2.0, 3.0, 4.0, 5.0]],
        ...     coords=[
        ...         ("area (ISO3)", ["COL", "MEX"]),
        ...         ("time", pd.date_range("2000", "2003", freq="YS")),
        ...     ],
        ... )
        >>> da
        <xarray.DataArray (area (ISO3): 2, time: 4)> Size: 64B
        array([[0., 1., 2., 3.],
               [2., 3., 4., 5.]])
        Coordinates:
          * area (ISO3)  (area (ISO3)) <U3 24B 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01

        Setting an existing value

        >>> da.pr.set("area", "COL", np.array([0.5, 0.6, 0.7, 0.8]))
        Traceback (most recent call last):
        ...
        ValueError: Values {'COL'} for 'area (ISO3)' already exist and contain data. ...
        >>> da.pr.set(
        ...     "area", "COL", np.array([0.5, 0.6, 0.7, 0.8]), existing="overwrite"
        ... )
        <xarray.DataArray (area (ISO3): 2, time: 4)> Size: 64B
        array([[0.5, 0.6, 0.7, 0.8],
               [2. , 3. , 4. , 5. ]])
        Coordinates:
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
          * area (ISO3)  (area (ISO3)) <U3 24B 'COL' 'MEX'

        By default, existing values are only overwritten if all existing values are
        NaN

        >>> da_partly_empty = da.copy(deep=True)
        >>> da_partly_empty.pr.loc[{"area": "COL"}] = np.nan
        >>> da_partly_empty
        <xarray.DataArray (area (ISO3): 2, time: 4)> Size: 64B
        array([[nan, nan, nan, nan],
               [ 2.,  3.,  4.,  5.]])
        Coordinates:
          * area (ISO3)  (area (ISO3)) <U3 24B 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        >>> da_partly_empty.pr.set("area", "COL", np.array([0.5, 0.6, 0.7, 0.8]))
        <xarray.DataArray (area (ISO3): 2, time: 4)> Size: 64B
        array([[0.5, 0.6, 0.7, 0.8],
               [2. , 3. , 4. , 5. ]])
        Coordinates:
          * area (ISO3)  (area (ISO3)) <U3 24B 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        >>> # if even one value contains data, the default is to raise an Error
        >>> da_partly_empty.pr.loc[{"area": "COL", "time": "2001"}] = 0.6
        >>> da_partly_empty
        <xarray.DataArray (area (ISO3): 2, time: 4)> Size: 64B
        array([[nan, 0.6, nan, nan],
               [2. , 3. , 4. , 5. ]])
        Coordinates:
          * area (ISO3)  (area (ISO3)) <U3 24B 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        >>> da_partly_empty.pr.set("area", "COL", np.array([0.5, 0.6, 0.7, 0.8]))
        Traceback (most recent call last):
        ...
        ValueError: Values {'COL'} for 'area (ISO3)' already exist and contain data. ...

        Introducing a new value uses the same syntax as modifying existing values

        >>> da.pr.set("area", "ARG", np.array([0.5, 0.6, 0.7, 0.8]))
        <xarray.DataArray (area (ISO3): 3, time: 4)> Size: 96B
        array([[0.5, 0.6, 0.7, 0.8],
               [0. , 1. , 2. , 3. ],
               [2. , 3. , 4. , 5. ]])
        Coordinates:
          * area (ISO3)  (area (ISO3)) <U3 36B 'ARG' 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01

        You can also mix existing and new values

        >>> da.pr.set(
        ...     "area",
        ...     ["COL", "ARG"],
        ...     np.array([[0.5, 0.6, 0.7, 0.8], [5, 6, 7, 8]]),
        ...     existing="overwrite",
        ... )
        <xarray.DataArray (area (ISO3): 3, time: 4)> Size: 96B
        array([[5. , 6. , 7. , 8. ],
               [0.5, 0.6, 0.7, 0.8],
               [2. , 3. , 4. , 5. ]])
        Coordinates:
          * area (ISO3)  (area (ISO3)) <U3 36B 'ARG' 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01

        If you don't want to automatically extend the dimensions with new values, you
        can request checking that all keys already exist using ``new="error"``:

        >>> da.pr.set("area", "ARG", np.array([0.5, 0.6, 0.7, 0.8]), new="error")
        Traceback (most recent call last):
        ...
        KeyError: "Values {'ARG'} not in 'area (ISO3)', use new='extend' to automatic...

        If you want to use broadcasting or have more dimensions, the dimensions of your
        input can't be determined automatically anymore. Use the value_dims parameter
        to supply this information.

        >>> da.pr.set(
        ...     "area",
        ...     ["COL", "ARG"],
        ...     np.array([0.5, 0.6, 0.7, 0.8]),
        ...     existing="overwrite",
        ... )
        Traceback (most recent call last):
        ...
        ValueError: Could not automatically determine value dimensions, please use th...
        >>> da.pr.set(
        ...     "area",
        ...     ["COL", "ARG"],
        ...     np.array([0.5, 0.6, 0.7, 0.8]),
        ...     value_dims=["time"],
        ...     existing="overwrite",
        ... )
        <xarray.DataArray (area (ISO3): 3, time: 4)> Size: 96B
        array([[0.5, 0.6, 0.7, 0.8],
               [0.5, 0.6, 0.7, 0.8],
               [2. , 3. , 4. , 5. ]])
        Coordinates:
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
          * area (ISO3)  (area (ISO3)) <U3 36B 'ARG' 'COL' 'MEX'

        Instead of overwriting existing values, you can also choose to only fill missing
        values.

        >>> da.pr.loc[{"area": "COL", "time": "2001"}] = np.nan
        >>> da
        <xarray.DataArray (area (ISO3): 2, time: 4)> Size: 64B
        array([[ 0., nan,  2.,  3.],
               [ 2.,  3.,  4.,  5.]])
        Coordinates:
          * area (ISO3)  (area (ISO3)) <U3 24B 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        >>> da.pr.set(
        ...     "area",
        ...     ["COL", "ARG"],
        ...     np.array([0.5, 0.6, 0.7, 0.8]),
        ...     value_dims=["time"],
        ...     existing="fillna",
        ... )
        <xarray.DataArray (area (ISO3): 3, time: 4)> Size: 96B
        array([[0.5, 0.6, 0.7, 0.8],
               [0. , 0.6, 2. , 3. ],
               [2. , 3. , 4. , 5. ]])
        Coordinates:
          * area (ISO3)  (area (ISO3)) <U3 36B 'ARG' 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01

        Because you can also supply a DataArray as a value, it is easy to define values
        from existing values using arithmetic

        >>> da.pr.set("area", "ARG", da.pr.loc[{"area": "COL"}] * 2)
        <xarray.DataArray (area (ISO3): 3, time: 4)> Size: 96B
        array([[ 0., nan,  4.,  6.],
               [ 0., nan,  2.,  3.],
               [ 2.,  3.,  4.,  5.]])
        Coordinates:
          * area (ISO3)  (area (ISO3)) object 24B 'ARG' 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01

        Returns
        -------
        da : xr.DataArray
            modified DataArray
        """
        if np.ndim(key) == 0:  # scalar
            key = [key]
        else:
            key = key

        if existing == "error":
            already_existing = set(self._da[dim].values).intersection(set(key))
            if already_existing:
                raise ValueError(
                    f"Values {already_existing!r} for {dim!r} already exist."
                    " Use existing='overwrite' or 'fillna' to avoid this error."
                )
            # without conflicting keys, the new keys are filled
            existing = "fillna"

        if new not in ("error", "extend"):
            raise ValueError(
                f"If given, 'new' must specify one of 'error' or 'extend', not {new!r}."
            )

        if isinstance(value, xr.DataArray):
            if value_dims is not None:
                raise ValueError("value_dims given, but value is already a DataArray.")

            # conform value to given dim: key
            if dim not in value.dims:
                if dim in value.coords:
                    value = value.reset_coords([dim], drop=True)
                value = value.expand_dims({dim: key})  # type: ignore
            else:
                value = value.loc[{dim: key}]

            # we broadcast value to full self._da, a possible optimization would be
            # to broadcast value only to sel, but would need more careful handling
            # later.
            if new == "extend":
                expanded, value = xr.broadcast(self._da, value)
            else:
                expanded = self._da
                value = value.broadcast_like(expanded)

            sel = self._sel_error(expanded, dim, key)

            value.attrs = self._da.attrs
            value.name = self._da.name

        else:
            # convert value to DataArray

            if new == "extend":
                new_index = list(self._da[dim].values)
                for item in key:
                    if item not in new_index:
                        new_index.append(item)
                new_index = np.array(new_index, dtype=self._da[dim].dtype)
                expanded = self._da.reindex({dim: new_index}, copy=False)
            else:
                expanded = self._da

            sel = self._sel_error(expanded, dim, key)

            if value_dims is None:
                value_dims = []
                for i, idim in enumerate(sel.dims):
                    if sel.shape[i] > 1:
                        value_dims.append(idim)
                if len(value_dims) != len(value.shape):
                    raise ValueError(
                        "Could not automatically determine value dimensions, please"
                        " use the value_dims parameter."
                    )
            value = xr.DataArray(
                value,
                coords=[(idim, sel[idim].data) for idim in value_dims],
                name=self._da.name,
                attrs=self._da.attrs,
            ).broadcast_like(sel)

        if existing == "fillna_empty":
            if sel.count().item() > 0:
                already_existing = set(self._da[dim].values).intersection(set(key))
                raise ValueError(
                    f"Values {already_existing!r} for {dim!r} already exist and contain"
                    f" data. Use existing='overwrite' or 'fillna' to avoid this error."
                )
            existing = "fillna"

        if existing == "overwrite":
            return value.combine_first(expanded)
        elif existing == "fillna":
            return expanded.combine_first(value)
        else:
            raise ValueError(
                "If given, 'existing' must specify one of 'error', 'overwrite', "
                f"'fillna_empty', or 'fillna', not {existing!r}."
            )


class DatasetSettersAccessor(_accessor_base.BaseDatasetAccessor):
    @staticmethod
    def _set_apply(
        da: xr.DataArray,
        dim: typing.Hashable,
        key: typing.Any,
        value: xr.Dataset,
        existing: str,
        new: str,
    ) -> xr.DataArray:
        if dim not in da.dims:
            return da
        return da.pr.set(
            dim=dim, key=key, value=value[da.name], existing=existing, new=new
        )

    @alias_dims(["dim"])
    def set(
        self,
        dim: typing.Hashable,
        key: typing.Any,
        value: xr.Dataset,
        *,
        existing: str = "fillna_empty",
        new: str = "extend",
    ) -> xr.Dataset:
        """Set values, optionally expanding the given dimension as necessary.

        All data variables which have the given dimension are modified.
        The affected data variables are mutated using
        ``DataArray.pr.set(dim, key, value[name], existing=existing, new=new)``.

        Parameters
        ----------
        dim: str
            Dimension along which values should be set. Only data variables which have
            this dimension are mutated.
        key: scalar or list of scalars
            Keys in the dimension which should be set. Key values which are missing
            in the dimension are inserted. The handling of key values which already
            exist in the dimension is determined by the ``existing`` parameter.
        value: xr.Dataset
            Values that will be inserted at the positions specified by ``key``.
            ``value`` needs to contain all data variables which have the dimension.
            ``value`` has to be broadcastable to ``ds.pr.loc[{dim: key}]``.
        existing: "fillna_empty", "error", "overwrite", or "fillna", optional
            How to handle existing keys. If ``existing="fillna_empty"`` (default), new
            values overwrite existing values only if all existing values are NaN.
            If ``existing="error"``, a ValueError is raised if any key already exists
            in the index. If ``existing="overwrite"``, new values overwrite current
            values for existing keys. If ``existing="fillna"``, the new values only
            overwrite NaN values for existing keys.
        new: "extend", or "error", optional
            How to handle new keys. If ``new="extend"`` (default), keys which do not
            exist so far are automatically inserted by extending the dimension.
            If ``new="error"``, a KeyError is raised if any key is not yet in the
            dimension.

        Examples
        --------
        >>> import pandas as pd
        >>> import xarray as xr
        >>> import numpy as np
        >>> area = ("area (ISO3)", ["COL", "MEX"])
        >>> time = ("time", pd.date_range("2000", "2003", freq="YS"))
        >>> ds = xr.Dataset(
        ...     {
        ...         "CO2": xr.DataArray(
        ...             [[0.0, 1.0, 2.0, 3.0], [2.0, 3.0, 4.0, 5.0]],
        ...             coords=[area, time],
        ...         ),
        ...         "SF4": xr.DataArray(
        ...             [[0.5, 1.5, 2.5, 3.5], [2.5, 3.5, np.nan, 5.5]],
        ...             coords=[area, time],
        ...         ),
        ...     },
        ...     attrs={"area": "area (ISO3)"},
        ... )
        >>> ds
        <xarray.Dataset> Size: ...
        Dimensions:      (area (ISO3): 2, time: 4)
        Coordinates:
          * area (ISO3)  (area (ISO3)) <U3 24B 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        Data variables:
            CO2          (area (ISO3), time) float64 64B 0.0 1.0 2.0 3.0 2.0 3.0 4.0 5.0
            SF4          (area (ISO3), time) float64 64B 0.5 1.5 2.5 3.5 2.5 3.5 nan 5.5
        Attributes:
            area:     area (ISO3)

        Setting an existing value

        >>> ds.pr.set("area", "MEX", ds.pr.loc[{"area": "COL"}] * 20)
        Traceback (most recent call last):
        ...
        ValueError: Values {'MEX'} for 'area (ISO3)' already exist and contain data. ...
        >>> ds.pr.set(
        ...     "area", "MEX", ds.pr.loc[{"area": "COL"}] * 20, existing="overwrite"
        ... )
        <xarray.Dataset> Size: ...
        Dimensions:      (area (ISO3): 2, time: 4)
        Coordinates:
          * area (ISO3)  (area (ISO3)) object 16B 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        Data variables:
            CO2          (area (ISO3), time) float64 64B 0.0 1.0 2.0 ... 20.0 40.0 60.0
            SF4          (area (ISO3), time) float64 64B 0.5 1.5 2.5 ... 30.0 50.0 70.0
        Attributes:
            area:     area (ISO3)

        Instead of overwriting existing values, you can also choose to only fill
        missing values

        >>> ds.pr.set("area", "MEX", ds.pr.loc[{"area": "COL"}] * 20, existing="fillna")
        <xarray.Dataset> Size: ...
        Dimensions:      (area (ISO3): 2, time: 4)
        Coordinates:
          * area (ISO3)  (area (ISO3)) object 16B 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        Data variables:
            CO2          (area (ISO3), time) float64 64B 0.0 1.0 2.0 3.0 2.0 3.0 4.0 5.0
            SF4          (area (ISO3), time) float64 64B 0.5 1.5 2.5 ... 3.5 50.0 5.5
        Attributes:
            area:     area (ISO3)

        By default, existing values are only filled if all existing values are
        missing in all data variables

        >>> ds_partly_empty = ds.copy(deep=True)
        >>> ds_partly_empty["CO2"].pr.loc[{"area": "COL"}] = np.nan
        >>> ds_partly_empty["SF4"].pr.loc[{"area": "COL"}] = np.nan
        >>> ds_partly_empty
        <xarray.Dataset> Size: ...
        Dimensions:      (area (ISO3): 2, time: 4)
        Coordinates:
          * area (ISO3)  (area (ISO3)) <U3 24B 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        Data variables:
            CO2          (area (ISO3), time) float64 64B nan nan nan nan 2.0 3.0 4.0 5.0
            SF4          (area (ISO3), time) float64 64B nan nan nan nan 2.5 3.5 nan 5.5
        Attributes:
            area:     area (ISO3)
        >>> ds_partly_empty.pr.set(
        ...     "area", "COL", ds_partly_empty.pr.loc[{"area": "MEX"}] * 10
        ... )
        <xarray.Dataset> Size: ...
        Dimensions:      (area (ISO3): 2, time: 4)
        Coordinates:
          * area (ISO3)  (area (ISO3)) object 16B 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        Data variables:
            CO2          (area (ISO3), time) float64 64B 20.0 30.0 40.0 ... 3.0 4.0 5.0
            SF4          (area (ISO3), time) float64 64B 25.0 35.0 nan ... 3.5 nan 5.5
        Attributes:
            area:     area (ISO3)
        >>> # if even one value is non-nan, this fails by default
        >>> ds_partly_empty["SF4"].pr.loc[{"area": "COL", "time": "2001"}] = 2
        >>> ds_partly_empty.pr.set(
        ...     "area", "COL", ds_partly_empty.pr.loc[{"area": "MEX"}] * 10
        ... )
        Traceback (most recent call last):
        ...
        ValueError: Values {'COL'} for 'area (ISO3)' already exist and contain data. ...

        Introducing a new value uses the same syntax

        >>> ds.pr.set("area", "BOL", ds.pr.loc[{"area": "COL"}] * 20)
        <xarray.Dataset> Size: ...
        Dimensions:      (area (ISO3): 3, time: 4)
        Coordinates:
          * area (ISO3)  (area (ISO3)) object 24B 'BOL' 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        Data variables:
            CO2          (area (ISO3), time) float64 96B 0.0 20.0 40.0 ... 3.0 4.0 5.0
            SF4          (area (ISO3), time) float64 96B 10.0 30.0 50.0 ... 3.5 nan 5.5
        Attributes:
            area:     area (ISO3)

        If you don't want to automatically extend the dimensions with new values, you
        can request checking that all keys already exist using ``new="error"``:

        >>> ds.pr.set("area", "BOL", ds.pr.loc[{"area": "COL"}] * 20, new="error")
        Traceback (most recent call last):
        ...
        KeyError: "Values {'BOL'} not in 'area (ISO3)', use new='extend' to automatic...

        Note that data variables which do not contain the specified dimension are
        ignored

        >>> ds["population"] = xr.DataArray([1e6, 1.2e6, 1.3e6, 1.4e6], coords=(time,))
        >>> ds
        <xarray.Dataset> Size: ...
        Dimensions:      (area (ISO3): 2, time: 4)
        Coordinates:
          * area (ISO3)  (area (ISO3)) <U3 24B 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        Data variables:
            CO2          (area (ISO3), time) float64 64B 0.0 1.0 2.0 3.0 2.0 3.0 4.0 5.0
            SF4          (area (ISO3), time) float64 64B 0.5 1.5 2.5 3.5 2.5 3.5 nan 5.5
            population   (time) float64 32B 1e+06 1.2e+06 1.3e+06 1.4e+06
        Attributes:
            area:     area (ISO3)
        >>> ds.pr.set("area", "BOL", ds.pr.loc[{"area": "COL"}] * 20)
        <xarray.Dataset> Size: ...
        Dimensions:      (area (ISO3): 3, time: 4)
        Coordinates:
          * area (ISO3)  (area (ISO3)) object 24B 'BOL' 'COL' 'MEX'
          * time         (time) datetime64[ns] 32B 2000-01-01 2001-01-01 ... 2003-01-01
        Data variables:
            CO2          (area (ISO3), time) float64 96B 0.0 20.0 40.0 ... 3.0 4.0 5.0
            SF4          (area (ISO3), time) float64 96B 10.0 30.0 50.0 ... 3.5 nan 5.5
            population   (time) float64 32B 1e+06 1.2e+06 1.3e+06 1.4e+06
        Attributes:
            area:     area (ISO3)

        Returns
        -------
        ds : xr.Dataset
            modified Dataset
        """
        if not isinstance(value, xr.Dataset):
            raise TypeError(f"value must be a Dataset, not {type(value)}")

        if self._ds.pr.has_processing_info():
            raise NotImplementedError(
                "Dataset contains processing information, this is not supported yet. "
                "Use ds.pr.remove_processing_info()."
            )

        return self._ds.map(
            self._set_apply,
            keep_attrs=True,
            dim=dim,
            key=key,
            value=value,
            existing=existing,
            new=new,
        )



#!/usr/bin/env python
"""Tests for _downscale.py"""

import numpy as np
import pytest
import xarray as xr

from primap2 import ureg

from .utils import allclose, assert_equal


def test_downscale_gas_timeseries(empty_ds):
    for key in empty_ds:
        empty_ds[key].pint.magnitude[:] = np.nan
    empty_ds["CO2"].loc[{"time": "2002"}] = 1 * ureg("Gg CO2 / year")
    empty_ds["SF6"].loc[{"time": "2002"}] = 1 * ureg("Gg SF6 / year")
    empty_ds["CH4"].loc[{"time": "2002"}] = 1 * ureg("Gg CH4 / year")
    sf6 = 22_800
    ch4 = 25
    empty_ds["KYOTOGHG (AR4GWP100)"][:] = (1 + sf6 + ch4) * ureg("Gg CO2 / year")
    empty_ds["KYOTOGHG (AR4GWP100)"].loc[{"time": "2020"}] = (
        2 * (1 + sf6 + ch4) * ureg("Gg CO2 / year")
    )

    downscaled = empty_ds.pr.downscale_gas_timeseries(
        basket="KYOTOGHG (AR4GWP100)", basket_contents=["CO2", "SF6", "CH4"]
    )
    expected = empty_ds.copy()
    expected["CO2"][:] = 1 * ureg("Gg CO2 / year")
    expected["SF6"][:] = 1 * ureg("Gg SF6 / year")
    expected["CH4"][:] = 1 * ureg("Gg CH4 / year")
    expected["CO2"].loc[{"time": "2020"}] = 2 * ureg("Gg CO2 / year")
    expected["SF6"].loc[{"time": "2020"}] = 2 * ureg("Gg SF6 / year")
    expected["CH4"].loc[{"time": "2020"}] = 2 * ureg("Gg CH4 / year")

    xr.testing.assert_identical(downscaled, expected)

    with pytest.raises(
        ValueError,
        match="Only one of 'skipna' and 'skipna_evaluation_dims' may be supplied, not both.",
    ):
        empty_ds.pr.downscale_gas_timeseries(
            basket="KYOTOGHG (AR4GWP100)",
            basket_contents=["CO2", "SF6", "CH4"],
            skipna_evaluation_dims=["time"],
            skipna=True,
        )

    empty_ds["SF6"].loc[{"time": "2002"}] = 2 * ureg("Gg SF6 / year")

    with pytest.raises(
        ValueError, match="To continue regardless, set check_consistency=False"
    ):
        empty_ds.pr.downscale_gas_timeseries(
            basket="KYOTOGHG (AR4GWP100)", basket_contents=["CO2", "SF6", "CH4"]
        )


def test_downscale_timeseries(empty_ds):
    for key in empty_ds:
        empty_ds[key].pint.magnitude[:] = np.nan
    t = empty_ds.loc[{"area (ISO3)": "BOL"}].copy()
    t["area (ISO3)"] = ["CAMB"]  # here, the sum of COL, ARG, MEX, and BOL
    ds = xr.concat([empty_ds, t], dim="area (ISO3)")
    da: xr.DataArray = ds["CO2"]

    da.loc[{"area (ISO3)": ["COL", "ARG", "MEX"], "time": "2002"}] = 1 * ureg(
        "Gg CO2 / year"
    )
    da.loc[{"area (ISO3)": "BOL", "time": "2002"}] = 3 * ureg("Gg CO2 / year")
    da.loc[{"area (ISO3)": "CAMB", "time": "2002"}] = 6 * ureg("Gg CO2 / year")

    da.loc[{"area (ISO3)": ["COL", "ARG", "MEX", "BOL"], "time": "2012"}] = 2 * ureg(
        "Gg CO2 / year"
    )

    da.loc[{"area (ISO3)": "CAMB", "source": "RAND2020"}] = np.concatenate(
        [np.array([6] * 11), np.stack([8, 8]), np.linspace(8, 10, 8)]
    ) * ureg("Gg CO2 / year")

    downscaled = da.pr.downscale_timeseries(
        dim="area (ISO3)", basket="CAMB", basket_contents=["COL", "ARG", "MEX", "BOL"]
    )
    expected = da.copy()

    expected.loc[{"area (ISO3)": ["COL", "ARG", "MEX"], "source": "RAND2020"}] = (
        np.broadcast_to(
            np.concatenate(
                [
                    np.array([1, 1]),
                    np.linspace(1 / 6, 2 / 8, 11) * np.array([6] * 9 + [8] * 2),
                    np.linspace(2, 2 * 10 / 8, 8),
                ]
            ),
            (3, 21),
        ).T
        * ureg("Gg CO2 / year")
    )
    expected.loc[{"area (ISO3)": "BOL", "source": "RAND2020"}] = np.concatenate(
        [
            np.array([3, 3]),
            np.linspace(3 / 6, 2 / 8, 11) * np.array([6] * 9 + [8] * 2),
            np.linspace(2, 2 * 10 / 8, 8),
        ]
    ) * ureg("Gg CO2 / year")

    # we need a higher atol, because downscale_timeseries actually does the
    # downscaling using a proper calendar while here we use a calendar where all years
    # have the same length.
    assert_equal(downscaled, expected, equal_nan=True, atol=0.01)
    allclose(
        downscaled.loc[{"area (ISO3)": "CAMB"}],
        downscaled.loc[{"area (ISO3)": ["COL", "ARG", "MEX", "BOL"]}].sum(
            dim="area (ISO3)"
        ),
    )

    downscaled_ds = ds.pr.downscale_timeseries(
        dim="area (ISO3)", basket="CAMB", basket_contents=["COL", "ARG", "MEX", "BOL"]
    )
    assert_equal(downscaled_ds["CO2"], expected, equal_nan=True, atol=0.01)

    with pytest.raises(
        ValueError,
        match="Only one of 'skipna' and 'skipna_evaluation_dims' may be supplied, not both.",
    ):
        ds.pr.downscale_timeseries(
            dim="area (ISO3)",
            basket="CAMB",
            basket_contents=["COL", "ARG", "MEX", "BOL"],
            skipna_evaluation_dims=["time"],
            skipna=True,
        )

    da.loc[{"area (ISO3)": "BOL", "time": "2002"}] = 2 * ureg("Gg CO2 / year")
    with pytest.raises(
        ValueError, match="To continue regardless, set check_consistency=False"
    ):
        da.pr.downscale_timeseries(
            dim="area (ISO3)",
            basket="CAMB",
            basket_contents=["COL", "ARG", "MEX", "BOL"],
        )

    downscaled = da.pr.downscale_timeseries(
        dim="area (ISO3)",
        basket="CAMB",
        basket_contents=["COL", "ARG", "MEX", "BOL"],
        check_consistency=False,
    )

    expected = da.copy()

    expected.loc[{"area (ISO3)": ["COL", "ARG", "MEX"], "source": "RAND2020"}] = (
        np.broadcast_to(
            np.concatenate(
                [
                    np.array([1.2, 1.2, 1]),
                    (np.linspace(1 / 5, 2 / 8, 11) * np.array([6] * 9 + [8] * 2))[1:],
                    np.linspace(2, 2 * 10 / 8, 8),
                ]
            ),
            (3, 21),
        ).T
        * ureg("Gg CO2 / year")
    )
    expected.loc[{"area (ISO3)": "BOL", "source": "RAND2020"}] = np.concatenate(
        [
            np.array([2.4, 2.4, 2]),
            (np.linspace(2 / 5, 2 / 8, 11) * np.array([6] * 9 + [8] * 2))[1:],
            np.linspace(2, 2 * 10 / 8, 8),
        ]
    ) * ureg("Gg CO2 / year")

    assert_equal(downscaled, expected, equal_nan=True, atol=0.01)

    downscaled = da.pr.downscale_timeseries(
        dim="area (ISO3)",
        basket="CAMB",
        basket_contents=["COL", "ARG", "MEX", "BOL"],
        check_consistency=False,
        sel={"time": slice("2005", "2020")},
    )
    expected = da.copy()

    expected.loc[
        {"area (ISO3)": ["COL", "ARG", "MEX", "BOL"], "source": "RAND2020"}
    ] = np.broadcast_to(
        np.concatenate(
            [
                np.array(
                    [
                        np.nan,
                        np.nan,
                        1,
                        np.nan,
                        np.nan,
                        6 / 4,
                        6 / 4,
                        6 / 4,
                        6 / 4,
                        6 / 4,
                        6 / 4,
                        2,
                        2,
                    ]
                ),
                np.linspace(2, 2 * 10 / 8, 8),
            ]
        ),
        (4, 21),
    ).T * ureg("Gg CO2 / year")
    expected.loc[{"area (ISO3)": "BOL", "time": "2002"}] = 2 * ureg("Gg CO2 / year")

    assert_equal(downscaled, expected, equal_nan=True, atol=0.01)



"""Check if the used version of Python is good enough for us."""

import itertools
import sys

SUPPORTED_MAJOR_VERSIONS = (3,)
SUPPORTED_MINOR_VERSIONS = (10, 11)

if (
    sys.version_info.major not in SUPPORTED_MAJOR_VERSIONS
    or sys.version_info.minor not in SUPPORTED_MINOR_VERSIONS
):
    supported_versions = itertools.product(
        SUPPORTED_MAJOR_VERSIONS, SUPPORTED_MINOR_VERSIONS
    )
    supported_versions_human_readable = ", ".join(
        ".".join(str(x) for x in version) for version in supported_versions
    )
    print(
        f"Python version {sys.version_info} not supported, please install Python"
        f" in one of the supported versions: {supported_versions_human_readable}."
    )
    sys.exit(1)



"""Tests for _metadata.py"""

import datetime


def test_metadata_properties(opulent_ds):
    ds = opulent_ds
    assert ds.pr.references == "doi:10.1012"
    assert ds.pr.rights == "Use however you want."
    assert ds.pr.contact == "lol_no_one_will_answer@example.com"
    assert ds.pr.title == "Completely invented GHG inventory data"
    assert ds.pr.comment == "GHG inventory data ..."
    assert ds.pr.institution == "PIK"
    assert ds.pr.entity_terminology == "primap2"
    assert ds.pr.publication_date == datetime.date(2099, 12, 31)

    ds.pr.references = "references"
    assert ds.pr.references == "references"
    ds.pr.rights = "rights"
    assert ds.pr.rights == "rights"
    ds.pr.contact = "contact"
    assert ds.pr.contact == "contact"
    ds.pr.title = "title"
    assert ds.pr.title == "title"
    ds.pr.comment = "comment"
    assert ds.pr.comment == "comment"
    ds.pr.institution = "institution"
    assert ds.pr.institution == "institution"
    ds.pr.entity_terminology = "entity_terminology"
    assert ds.pr.entity_terminology == "entity_terminology"
    today = datetime.date.today()
    ds.pr.publication_date = today
    assert ds.pr.publication_date == today



"""Base classes for accessor mix-in classes."""

import typing

import xarray as xr

XrObj = typing.TypeVar("XrObj", xr.Dataset, xr.Dataset)


class BaseDataArrayAccessor:
    def __init__(self, da: xr.DataArray):
        self._da = da


class BaseDatasetAccessor:
    def __init__(self, ds: xr.Dataset):
        self._ds = ds
